'''
PipedViewerPQ is a graphics viewer application written in PyQt
that receives its drawing and other commands primarily from another
application through a pipe.  A limited number of commands are
provided by the viewer itself to allow saving and some manipulation
of the displayed image.  The controlling application, however, may
be unaware of these modifications made to the image.

PipedViewerPQProcess is used to create and run a PipedViewerPQ.

This package was developed by the Thermal Modeling and Analysis
Project (TMAP) of the National Oceanographic and Atmospheric
Administration's (NOAA) Pacific Marine Environmental Lab (PMEL).
'''

from __future__ import print_function

import sys
import os
import signal
import time
import math

# First try to import PySide2, then try PyQt5 if that fails, and finally try PyQt4 if that fails
try:
    import PySide2
    PYTHONQT_VERSION = 'PySide2'
except ImportError:
    try:
        import PyQt5
        PYTHONQT_VERSION = 'PyQt5'
    except ImportError:
        import PyQt4
        PYTHONQT_VERSION = 'PyQt4'

# Now that the Python Qt version is determined, import the parts
# allowing any import errors to propagate out
if PYTHONQT_VERSION == 'PySide2':
    from PySide2.QtCore    import Qt, QPointF, QRect, QRectF, QSize, QSizeF, QTimer
    from PySide2.QtGui     import QBrush, QColor, QFontMetricsF, QImage, QPainter, \
                                  QPalette, QPen, QPicture, QPixmap, QPolygonF, \
                                  QTextDocument
    from PySide2.QtWidgets import QAction, QApplication, QDialog, QFileDialog, QLabel, \
                                  QMainWindow, QMessageBox, QPushButton, QScrollArea
    from PySide2.QtSvg     import QSvgGenerator
    from PySide2.QtPrintSupport import QPrinter
elif PYTHONQT_VERSION == 'PyQt5':
    from PyQt5.QtCore    import Qt, QPointF, QRect, QRectF, QSize, QSizeF, QTimer
    from PyQt5.QtGui     import QBrush, QColor, QFontMetricsF, QImage, QPainter, \
                                QPalette, QPen, QPicture, QPixmap, QPolygonF, \
                                QTextDocument
    from PyQt5.QtWidgets import QAction, QApplication, QDialog, QFileDialog, QLabel, \
                                QMainWindow, QMessageBox, QPushButton, QScrollArea
    from PyQt5.QtSvg     import QSvgGenerator
    from PyQt5.QtPrintSupport import QPrinter
else:
    from PyQt4.QtCore import Qt, QPointF, QRect, QRectF, QSize, QSizeF, QTimer, QString
    from PyQt4.QtGui  import QAction, QApplication, QBrush, QColor, QDialog, \
                             QFileDialog, QFontMetricsF, QImage, QLabel, \
                             QMainWindow, QMessageBox, QPainter, QPalette, \
                             QPen, QPicture, QPixmap, QPolygonF, QPrinter, \
                             QPushButton, QScrollArea, QTextDocument
    from PyQt4.QtSvg  import QSvgGenerator

from multiprocessing import Pipe, Process

from pipedviewer import WINDOW_CLOSED_MESSAGE
from pipedviewer.cmndhelperpq import CmndHelperPQ
from pipedviewer.scaledialogpq import ScaleDialogPQ


class PipedViewerPQ(QMainWindow):
    '''
    A PyQt graphics viewer that receives generic drawing commands
    through a pipe.  Uses a list of QPictures to record the drawings
    which are then used to display, manipulate, and save the image.

    A drawing command is a dictionary with string keys that will be
    interpreted into the appropriate PyQt command(s).  For example,
        { "action":"drawText",
          "text":"Hello",
          "font":{"family":"Times", "size":100, "italic":True},
          "fill":{"color":0x880000, "style":"cross"},
          "outline":{"color":"black"},
          "location":(250,350) }

    The command { "action":"exit" } will shutdown the viewer.
    '''

    def __init__(self, cmndpipe, rspdpipe):
        '''
        Create a PyQt viewer which reads commands from the Pipe
        cmndpipe and writes responses back to rspdpipe.
        '''
        super(PipedViewerPQ, self).__init__()
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        # ignore Ctrl-C
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # default scene size
        self.__scenewidth = int(10.8 * self.physicalDpiX())
        self.__sceneheight = int(8.8 * self.physicalDpiY())
        # scaling factor for line widths and symbol sizes
        self.__widthfactor = None
        self.setWidthScalingFactor(0.72)
        # by default pay attention to any alpha channel values in colors
        self.__noalpha = False
        # initial default color for the background (opaque white)
        self.__lastclearcolor = QColor(0xFFFFFF)
        self.__lastclearcolor.setAlpha(0xFF)
        # List of QPictures creating the current scene
        self.__viewpics = [ ]
        self.__segid = None
        # QPicture/QPainter pair for the current view
        self.__activepicture = None
        self.__activepainter = None
        # Antialias when drawing?
        self.__antialias = True
        # data for recreating the current view
        self.__fracsides = None
        self.__clipit = True
        # number of drawing commands in the active painter
        self.__drawcount = 0
        # Limit the number of drawing commands per picture
        # to avoid the appearance of being "stuck"
        self.__maxdraws = 1024
        # scaling factor for creating the displayed scene
        self.__scalefactor = 1.0
        # automatically adjust the scaling factor to fit the window frame?
        self.__autoscale = True
        # values used to decide if the scene needs to be updated
        self.__lastpicdrawn = 0
        self.__createpixmap = True
        self.__clearpixmap = True
        # Calculations of modified rectangular regions in QPictures
        # currently do not account for width and height of QPictures
        # played inside them.  So keep a expansion value.
        self.__maxsymbolwidth = 0.0
        self.__maxsymbolheight = 0.0
        # create the label, that will serve as the canvas, in a scrolled area
        self.__scrollarea = QScrollArea(self)
        self.__label = QLabel(self.__scrollarea)
        # set the initial label size and other values for the scrolled area
        self.__label.setMinimumSize(self.__scenewidth, self.__sceneheight)
        self.__label.resize(self.__scenewidth, self.__sceneheight)
        # setup the scrolled area
        self.__scrollarea.setWidget(self.__label)
        self.__scrollarea.setBackgroundRole(QPalette.Dark)
        self.setCentralWidget(self.__scrollarea)
        self.__minsize = 128
        # default file name and format for saving the image
        self.__lastfilename = "ferret.png"
        self.__lastformat = "png"
        self.__addedannomargin = 12
        # command helper object
        self.__helper = CmndHelperPQ(self)
        # Create the menubar
        self.__scaleact = QAction(self.tr("&Scale"), self,
                                shortcut=self.tr("Ctrl+S"),
                                statusTip=self.tr("Scale the image (canvas and image change size)"),
                                triggered=self.inquireSceneScale)
        self.__saveact = QAction(self.tr("Save &As..."), self,
                                shortcut=self.tr("Ctrl+A"),
                                statusTip=self.tr("Save the image to file"),
                                triggered=self.inquireSaveFilename)
        self.__redrawact = QAction(self.tr("&Redraw"), self,
                                shortcut=self.tr("Ctrl+R"),
                                statusTip=self.tr("Clear and redraw the image"),
                                triggered=self.redrawScene)
        self.__aboutact = QAction(self.tr("&About"), self,
                                statusTip=self.tr("Show information about this viewer"),
                                triggered=self.aboutMsg)
        self.__aboutqtact = QAction(self.tr("About &Qt"), self,
                                statusTip=self.tr("Show information about the Qt library"),
                                triggered=self.aboutQtMsg)
        self.createMenus()
        # Set the initial size of the viewer
        self.__framedelta = 4
        mwwidth = self.__scenewidth + self.__framedelta
        mwheight = self.__sceneheight + self.__framedelta \
                 + self.menuBar().height() \
                 + self.statusBar().height()
        self.resize(mwwidth, mwheight)
        # check the command queue any time there are no window events to deal with
        self.__timer = QTimer(self)
        self.__timer.timeout.connect(self.checkCommandPipe)
        self.__timer.setInterval(0)
        self.__timer.start()
        # initialize the parameters for watermark image display
        self.__wmarkFilename = None
        self.__xloc = None
        self.__yloc = None
        self.__scalefrac = None
        self.__opacity = None
        self.__wmkdrawn = False

    def createMenus(self):
        '''
        Create the menu items for the viewer
        using the previously created actions.
        '''
        menuBar = self.menuBar()
        sceneMenu = menuBar.addMenu(menuBar.tr("&Image"))
        sceneMenu.addAction(self.__scaleact)
        sceneMenu.addAction(self.__saveact)
        sceneMenu.addAction(self.__redrawact)
        helpMenu = menuBar.addMenu(menuBar.tr("&Help"))
        helpMenu.addAction(self.__aboutact)
        helpMenu.addAction(self.__aboutqtact)

    def showEvent(self, event):
        '''
        When the viewer is going to be shown, make sure all
        the current pictures are displayed in the scene.
        '''
        # update, ignoring the visibility flags
        self.drawLastPictures(True)
        event.accept()

    def resizeEvent(self, event):
        '''
        Monitor resizing in case auto-scaling of the image is selected.
        '''
        if self.__autoscale:
            if self.autoScaleScene():
                # continue with the window resize
                event.accept()
            else:
                # another resize coming in, so ignore this one
                event.ignore()
        else:
            # continue with the window resize
            event.accept()

    def closeEvent(self, event):
        '''
        Clean up and send the WINDOW_CLOSED_MESSAGE on the response pipe
        before closing the window.
        '''
        self.__timer.stop()
        self.__cmndpipe.close()
        try:
            try:
                self.__rspdpipe.send(WINDOW_CLOSED_MESSAGE)
            finally:
                self.__rspdpipe.close()
        except Exception:
            pass
        event.accept()

    def exitViewer(self):
        '''
        Close and exit the viewer.
        '''
        self.close()

    def aboutMsg(self):
        QMessageBox.about(self, self.tr("About PipedViewerPQ"),
            self.tr("\n" \
            "PipedViewerPQ is a graphics viewer application that receives its " \
            "drawing and other commands primarily from another application " \
            "through a pipe.  A limited number of commands are provided by " \
            "the viewer itself to allow saving and some manipulation of the " \
            "displayed image.  The controlling application, however, may be " \
            "unaware of these modifications made to the image. " \
            "\n\n" \
            "PipedViewerPQ was developed by the Thermal Modeling and Analysis " \
            "Project (TMAP) of the National Oceanographic and Atmospheric " \
            "Administration's (NOAA) Pacific Marine Environmental Lab (PMEL). "))

    def aboutQtMsg(self):
        QMessageBox.aboutQt(self, self.tr("About Qt"))

    def ignoreAlpha(self):
        '''
        Return whether the alpha channel in colors should always be ignored.
        '''
        return self.__noalpha

    def paintScene(self, painter, first, leftx, uppery, scalefactor,
                   statusmsg, returnregion):
        '''
        Draws the pictures self.__viewpics[first:] using the QPainter
        painter.  This QPainter should have been initialized
        appropriately for the QPaintDevice to which it is painting
        (e.g., QImage.fill with the desired background color).

        The point (leftx, uppery) is the offset of the origin after
        scaling using scalefactor.  (All are floating point values.)

        The status bar will be updated with a message derived from
        statusmsg before drawing each picture.  Upon completion, the
        status bar will be cleared.

        If returnregion is True, a list of QRect objects describing
        the modified regions will be computed and returned.  If
        returnregion is False, the modified region will not be computed
        and an empty list will be returned.

        The call to painter.end() will need to be made after calling
        this function.
        '''
        # change the cursor to warn the user this may take some time
        QApplication.setOverrideCursor(Qt.WaitCursor)
        # create the incomplete status message
        if (first + 1) < len(self.__viewpics):
            mymsg = "%s (piece %%s of %s)" % (statusmsg, str(len(self.__viewpics)))
        else:
            mymsg = "%s (piece %%s)" % statusmsg
        # get the origin for drawing the pictures after scaling
        myorigin = QPointF(leftx, uppery)
        # set the scaling factor for the pictures
        painter.scale(scalefactor, scalefactor)
        modrects = [ ]
        # draw the appropriate pictures
        k = first
        for (viewpic, _) in self.__viewpics[first:]:
            k += 1
            # show the progress message
            self.statusBar().showMessage( self.tr(mymsg % str(k)) )
            # draw the picture
            painter.drawPicture(myorigin, viewpic)
            if returnregion:
                picrect = viewpic.boundingRect()
                if picrect.isValid():
                    # Expand the region to account for possible symbols
                    xval = picrect.x() - 0.5 * self.__maxsymbolwidth
                    yval = picrect.y() - 0.5 * self.__maxsymbolheight
                    width = picrect.width() + self.__maxsymbolwidth
                    height = picrect.height() + self.__maxsymbolheight
                    # Scale and translate the region, then convert to integer
                    xval = int( math.floor(xval * scalefactor + leftx) )
                    width = int( math.ceil(width * scalefactor) )
                    yval = int( math.floor(yval * scalefactor + uppery) )
                    height = int( math.ceil(height * scalefactor) )
                    # Add this rectangle to the list
                    modrects.append( QRect(xval, yval, width, height) )
        # done - clear the status message
        self.statusBar().clearMessage()
        # restore the cursor back to normal
        QApplication.restoreOverrideCursor()
        if (self.__wmarkFilename is not None) and not self.__wmkdrawn:
            # painter = QPainter(self.__label.pixmap())
            # Initialize watermark objects
            wmkpic = QPixmap(self.__wmarkFilename)
            wmkpt = QPointF()
            wmkpt.setX(self.__xloc)
            wmkpt.setY(self.__yloc)
            # set watermark image display opacity
            painter.setOpacity(self.__opacity)
            # set image scale
            painter.scale(self.__scalefrac, self.__scalefrac)
            # paint watermark image at specified location
            painter.setRenderHint(QPainter.Antialiasing)
            painter.drawPixmap(wmkpt, wmkpic)
            # painter.end()
            self.__wmkdrawn = True
        return modrects

    def drawLastPictures(self, ignorevis):
        '''
        Update the scene with pictures yet to be drawn.
        If ignorevis is True, the update will be done
        even if the viewer is not visible; otherwise
        drawing to the scene label is only done if the
        viewer is visible.
        '''
        if not ignorevis:
            if self.isMinimized() or not self.isVisible():
                # Not shown, so do not waste time drawing
                return
        if self.__createpixmap:
            # Create and assign a cleared pixmap
            mypixmap = QPixmap(self.__label.size())
            mypixmap.fill(self.__lastclearcolor)
            self.__label.setPixmap(mypixmap)
            self.__createpixmap = False
            self.__clearpixmap = False
            wascleared = True
        elif self.__clearpixmap:
            # Clear the existing pixmap
            self.__label.pixmap().fill(self.__lastclearcolor)
            self.__clearpixmap = False
            wascleared = True
        elif len(self.__viewpics) > self.__lastpicdrawn:
            # New pictures to add to an existing scene
            # wascleared = False
            # Rectangles of modified regins incorrect for drawText
            # so always update the entire scene
            wascleared = True
        else:
            # Nothing changed so just return
            return
        # Only create the QPainter if there are pictures
        # to draw (this is more than just a clear)
        if len(self.__viewpics) > self.__lastpicdrawn:
            painter = QPainter(self.__label.pixmap())
            modrects = self.paintScene(painter, self.__lastpicdrawn, \
                                       0.0, 0.0, self.__scalefactor, \
                                       "Drawing", not wascleared)
            painter.end()
        # if watermark is specified, display after other plotting occurs
        # if (self.__wmarkFilename is not None) and not self.__wmkdrawn:
        #     painter = QPainter(self.__label.pixmap())
        #     # Initialize watermark objects
        #     wmkpic = QPixmap(self.__wmarkFilename)
        #     wmkpt = QPointF()
        #     wmkpt.setX(self.__xloc)
        #     wmkpt.setY(self.__yloc)
        #     # set watermark image display opacity
        #     painter.setOpacity(self.__opacity)
        #     # set image scale
        #     painter.scale(self.__scalefrac, self.__scalefrac)
        #     # paint watermark image at specified location
        #     painter.setRenderHint(QPainter.Antialiasing)
        #     painter.drawPixmap(wmkpt, wmkpic)
        #     painter.end()
        #     self.__wmkdrawn = True
        # Notify the label of changes to the scene
        if wascleared:
            # the entire scene changed
            self.__label.update()
        else:
            # the scene changed only in the modrects areas
            for rect in modrects:
                self.__label.update(rect)
        # Update the record of which pictures have been displayed
        self.__lastpicdrawn = len(self.__viewpics)

    def clearScene(self, bkgcolor):
        '''
        Removes all view pictures, and fills the scene with bkgcolor.
        If bkgcolor is None or an invalid color, the color used is
        the one used from the last clearScene or redrawScene call
        with a valid color (or opaque white if a color has never
        been specified).
        '''
        # If there is an active View with content,
        # end it now, but do not update the scene
        if self.__activepainter and (self.__drawcount > 0):
            self.endView(False)
            restartview = True
        else:
            restartview = False
        # get the color to use for clearing (the background color)
        if bkgcolor:
            if bkgcolor.isValid():
                self.__lastclearcolor = bkgcolor
        # Delete all the pictures from the list and
        # mark that the pixmap needs to be cleared
        self.__viewpics[:] = [ ]
        self.__maxsymbolwidth = 0.0
        self.__maxsymbolheight = 0.0
        self.__clearpixmap = True
        self.__lastpicdrawn = 0
        # Update the scene label if visible
        self.drawLastPictures(False)
        # If there was an non-empty active View, restart it
        if restartview:
            self.beginViewFromSides(self.__fracsides, self.__clipit)

    def redrawScene(self, bkgcolor=None):
        '''
        Clear the scene using the given background color and redraw all
        the pictures to the displayed scene.  If bkgcolor is None or an
        invalid color, the color used is the one used from the last
        clearScene or redrawScene call with a valid color (or opaque
        white if a color has never been specified).
        '''
        # If there is an active View, end it now, but do not update the scene
        if self.__activepainter:
            self.endView(False)
            hadactiveview = True
        else:
            hadactiveview = False
        if bkgcolor:
            if bkgcolor.isValid():
                self.__lastclearcolor = bkgcolor
        # mark that the pixmap needs to be cleared
        # and all the pictures redrawn
        self.__clearpixmap = True
        self.__lastpicdrawn = 0
        # Update the scene label if visible
        self.drawLastPictures(False)
        # If there was an active View, restart it in this new system
        if hadactiveview:
            self.beginViewFromSides(self.__fracsides, self.__clipit)

    def resizeScene(self, width, height):
        '''
        Resize the scene to the given width and height in units of pixels.
        '''
        newwidth = int(width + 0.5)
        if newwidth < self.__minsize:
            newwidth = self.__minsize
        newheight = int(height + 0.5)
        if newheight < self.__minsize:
            newheight = self.__minsize
        if (newwidth != self.__scenewidth) or (newheight != self.__sceneheight):
            # Resize the label and set label values
            # so the scrollarea knows of the new size
            labelwidth = int(newwidth * self.__scalefactor + 0.5)
            labelheight = int(newheight * self.__scalefactor + 0.5)
            self.__label.setMinimumSize(labelwidth, labelheight)
            self.__label.resize(labelwidth, labelheight)
            # mark that the pixmap needs to be recreated
            self.__scenewidth = newwidth
            self.__sceneheight = newheight
            self.__createpixmap = True
            # If auto-scaling, set scaling factor to 1.0 and resize the window
            if self.__autoscale:
                self.__scalefactor = 1.0
                barheights = self.menuBar().height() + self.statusBar().height()
                self.resize(newwidth + self.__framedelta,
                            newheight + self.__framedelta + barheights)
                # the resize should redraw the scene
            else:
                # Redraw the scene from the beginning using the scaling factor
                self.redrawScene()


    def inquireSceneScale(self):
        '''
        Prompt the user for the desired scaling factor for the scene.
        '''
        labelwidth = int(self.__scenewidth * self.__scalefactor + 0.5)
        labelheight = int(self.__sceneheight * self.__scalefactor + 0.5)
        scaledlg = ScaleDialogPQ(self.__scalefactor, labelwidth, labelheight,
                       self.__minsize, self.__minsize, self.__autoscale, self)
        if scaledlg.exec_():
            (newscale, autoscale, okay) = scaledlg.getValues()
            if okay:
                if autoscale:
                    self.__autoscale = True
                    self.autoScaleScene()
                else:
                    self.__autoscale = False
                    self.scaleScene(newscale, False)

    def autoScaleScene(self):
        '''
        Selects a scaling factor that maximizes the scene within the window
        frame without requiring scroll bars.  Intended to be called when
        the window size is changed by the user and auto-scaling is turn on.

        Returns:
            True if scaling of this scene is done (no window resize)
            False if the a new resize command was issued
        '''
        barheights = self.menuBar().height() + self.statusBar().height()

        # get the size for the central widget
        cwheight = self.height() - barheights - self.__framedelta
        heightsf = float(cwheight) / float(self.__sceneheight)

        cwwidth = self.width() - self.__framedelta
        widthsf = float(cwwidth) / float(self.__scenewidth)

        if heightsf < widthsf:
            factor = heightsf
        else:
            factor = widthsf

        newcwheight = int(factor * self.__sceneheight + 0.5)
        newcwwidth = int(factor * self.__scenewidth + 0.5)

        # if the window does not have the correct aspect ratio, resize it so
        # it will; this will generate another call to this method.  Otherwise,
        # scale the scene and be done.
        if self.isMaximized() or \
           ( (abs(cwheight - newcwheight) <= self.__framedelta) and \
             (abs(cwwidth - newcwwidth) <= self.__framedelta) ):
            self.scaleScene(factor, False)
            return True
        else:
            self.resize(newcwwidth + self.__framedelta,
                        newcwheight + self.__framedelta + barheights)
            return False

    def scaleScene(self, factor, resizewin):
        '''
        Scales both the horizontal and vertical directions by factor.
        Scaling factors are not accumulative.  So if the scene was
        already scaled, that scaling is "removed" before this scaling
        factor is applied.  If resizewin is True, the main window is
        resized to accommodate this new scaled scene size.

        If factor is zero, just switch to auto-scaling at the current
        window size.  If factor is negative, rescale using the absolute
        value (possibly resizing the window) then switch to auto-scaling.
        '''
        fltfactor = float(factor)
        if fltfactor != 0.0:
            if resizewin:
                # from command - turn off autoscaling for the following
                # then turn back on if appropriate
                self.__autoscale = False
            newfactor = abs(fltfactor)
            newlabwidth = int(newfactor * self.__scenewidth + 0.5)
            newlabheight = int(newfactor * self.__sceneheight + 0.5)
            if (newlabwidth < self.__minsize) or (newlabheight < self.__minsize):
                # Set to minimum size
                if self.__scenewidth <= self.__sceneheight:
                    newfactor = float(self.__minsize) / float(self.__scenewidth)
                else:
                    newfactor = float(self.__minsize) / float(self.__sceneheight)
                newlabwidth = int(newfactor * self.__scenewidth + 0.5)
                newlabheight = int(newfactor * self.__sceneheight + 0.5)
            oldlabwidth = int(self.__scalefactor * self.__scenewidth + 0.5)
            oldlabheight = int(self.__scalefactor * self.__sceneheight + 0.5)
            if (newlabwidth != oldlabwidth) or (newlabheight != oldlabheight):
                # Set the new scaling factor
                self.__scalefactor = newfactor
                # Resize the label and set label values
                # so the scrollarea knows of the new size
                self.__label.setMinimumSize(newlabwidth, newlabheight)
                self.__label.resize(newlabwidth, newlabheight)
                # mark that the pixmap needs to be recreated
                self.__createpixmap = True
                # Redraw the scene from the beginning
                self.redrawScene()
            if resizewin:
                # resize the main window (if possible)
                barheights = self.menuBar().height() + self.statusBar().height()
                mwheight = newlabheight + barheights + self.__framedelta
                mwwidth = newlabwidth + self.__framedelta
                # Do not exceed the available real estate on the screen.
                # If autoscaling is in effect, the resize will trigger
                # any required adjustments.
                scrnrect = QApplication.desktop().availableGeometry()
                if mwwidth > 0.95 * scrnrect.width():
                    mwwidth = int(0.9 * scrnrect.width())
                if mwheight > 0.95 * scrnrect.height():
                    mwheight = int(0.9 * scrnrect.height())
                self.resize(mwwidth, mwheight)
        if fltfactor <= 0.0:
            # From command - turn on autoscaling
            self.__autoscale = True
            self.autoScaleScene();

    def inquireSaveFilename(self):
        '''
        Prompt the user for the name of the file into which to save the scene.
        The file format will be determined from the filename extension.
        '''
        formattypes = [ ( "png",
                          "PNG - Portable Networks Graphics (*.png)"),
                        ( "jpeg",
                          "JPEG - Joint Photographic Experts Group (*.jpeg *.jpg *.jpe)" ),
                        ( "tiff",
                          "TIFF - Tagged Image File Format (*.tiff *.tif)" ),
                        ( "pdf",
                          "PDF - Portable Document Format (*.pdf)" ),
                        ( "ps",
                          "PS - PostScript (*.ps)" ),
                        ( "svg",
                          "SVG - Scalable Vector Graphics (*.svg)" ),
                        ( "bmp",
                          "BMP - Windows Bitmap (*.bmp)" ),
                        ( "ppm",
                          "PPM - Portable Pixmap (*.ppm)" ),
                        ( "xpm",
                          "XPM - X11 Pixmap (*.xpm)" ),
                        ( "xbm",
                          "XBM - X11 Bitmap (*.xbm)" ), ]
        filters = ";;".join( [ t[1] for t in formattypes ] )
        if PYTHONQT_VERSION == 'PyQt4':
            # tr returns QStrings in PyQt4 (Python2)
            (fileName, fileFilter) = QFileDialog.getSaveFileNameAndFilter(self,
                 self.tr("Save the current image as "), self.tr(self.__lastfilename), self.tr(filters))
        else:
            # tr returns Python unicode strings in PySide2 or PyQt5 (Python3)
            (fileName, fileFilter) = QFileDialog.getSaveFileName(self,
                self.tr("Save the current image as "), self.tr(self.__lastfilename), self.tr(filters))
        if fileName:
            for (fmt, fmtQName) in formattypes:
                if self.tr(fmtQName) == fileFilter:
                    fileFormat = fmt
                    break
            else:
                raise RuntimeError("Unexpected file format name '%s'" % fileFilter)
            self.saveSceneToFile(fileName, fileFormat, None, None, None, None)
            self.__lastfilename = fileName
            self.__lastformat = fileFormat

    def saveSceneToFile(self, filename, imageformat, transparent,
                        vectsize, rastsize, myannotations):
        '''
        Save the current scene to the named file.  If imageformat
        is empty or None, the format is guessed from the filename
        extension.

        If transparent is False, the entire scene is initialized
        to the last clearing color used, using a filled rectangle
        for vector images.

        If given, vectsize is the size in inches of a saved vector
        image.  If vectsize is not given, a vector image will be
        saved at the current displayed scaled image size, unless
        specified otherwise if showPrintDialog is True.

        If given, rastsize is the pixels size of a saved raster
        image.  If rastsize is not given, a raster image will be
        saved at the current displayed scaled image size.

        If myannotations is not None, the strings given in the tuple
        are to be displayed above the image.  These annotations add
        height, as needed, to the saved image (i.e., vectsize or
        rastsize gives the height of the image below these annotations).
        '''
        # This could be called when there is no scene present.
        # If this is the case, ignore the call.
        if len(self.__viewpics) == 0:
            return
        if not imageformat:
            # Guess the image format from the filename extension
            # to determine if it is a vector type, and if so,
            # which type. All the raster types use a QImage, which
            # does this guessing of format when saving the image.
            fileext = ( os.path.splitext(filename)[1] ).lower()
            if fileext == '.pdf':
                # needs a PDF QPrinter
                myformat = 'pdf'
            elif fileext == '.eps':
                # needs a PS QPrinter and never rotate
                myformat = 'eps'
            elif fileext == '.ps':
                # needs a PS QPrinter
                myformat = 'ps'
            elif fileext == '.svg':
                # needs a QSvgGenerator
                myformat = 'svg'
            elif fileext == '.plt':
                # check for plt (gks metafile) - needs to be changed to pdf
                myformat = 'plt'
            elif fileext == '.gif':
                # check for gif - needs to be changed to png
                myformat = 'gif'
            else:
                # use a QImage and let it figure out the format
                myformat = None
        else:
            myformat = imageformat.lower()

        if myformat == 'plt':
            # Silently convert plt filename and format to pdf
            myformat = 'pdf'
            myfilename = os.path.splitext(filename)[0] + ".pdf"
        elif myformat == 'gif':
            # Silently convert gif filename and format to png
            myformat = 'png'
            myfilename = os.path.splitext(filename)[0] + ".png"
        else:
            myfilename = filename

        if myannotations:
            annopicture = QPicture()
            annopainter = QPainter(annopicture)
            annotextdoc = QTextDocument()
            # Leave room for the added margins to the width
            annotextdoc.setTextWidth(self.__scenewidth - 2.0 * self.__addedannomargin)
            annotextdoc.setHtml("<p>" + "<br />".join(myannotations) + "</p>")
            annotextdoc.drawContents(annopainter)
            annopainter.end()
            annosize = annotextdoc.documentLayout().documentSize()
        else:
            annopicture = None
            annosize = None

        if (myformat == 'ps') or (myformat == 'eps') or (myformat == 'pdf'):
            # Setup the QPrinter that will be used to create the EPS, PS, or PDF file
            printer = QPrinter(QPrinter.HighResolution)
            printer.setOutputFileName(myfilename)
            # The print format is automatically set from the
            # filename extension; so the following is actually
            # only needed for absent or strange extensions
            if (myformat == 'ps') or (myformat == 'eps'):
                printer.setOutputFormat(QPrinter.PostScriptFormat)
            else:
                printer.setOutputFormat(QPrinter.PdfFormat)
            # Print to file in color
            printer.setColorMode(printer.Color)
            # get the width and height in inches of the image to be produced
            if vectsize:
                imagewidth = vectsize.width()
                imageheight = vectsize.height()
            else:
                imagewidth = self.__scenewidth * self.__scalefactor \
                             / float(self.physicalDpiX())
                imageheight = self.__sceneheight * self.__scalefactor \
                              / float(self.physicalDpiY())
            # Add in any height needed for the annotations
            if annopicture:
                annoheight = (annosize.height() + 2 * self.__addedannomargin) * \
                             imageheight / self.__sceneheight
                imageheight += annoheight
            # Set the image size
            try:
                # Set custom paper size to just fit around the image
                if (myformat != 'eps') and (imagewidth > imageheight):
                    printer.setPaperSize(QSizeF(imageheight, imagewidth), QPrinter.Inch)
                else:
                    printer.setPaperSize(QSizeF(imagewidth, imageheight), QPrinter.Inch)
                # The above has issues with Qt 4.6 at GFDL -
                # still puts it on the default letter size page.
                # So just always use a letter size page.
                # printer.setPaperSize(QPrinter.Letter)
            except AttributeError:
                # setPaperSize introduced in 4.4 and made setPageSize
                # obsolete; but RHEL5 Qt4 is 4.2, so set to letter size
                printer.setPageSize(QPrinter.Letter)
            # No margins (setPageMargins introduced in 4.4)
            printer.setFullPage(True)
            # Default orientation
            if (myformat != 'eps') and (imagewidth > imageheight):
                printer.setOrientation(QPrinter.Landscape)
            else:
                printer.setOrientation(QPrinter.Portrait)
            # also get the image size in units of printer dots
            printres = printer.resolution()
            printwidth = int(imagewidth * printres + 0.5)
            printheight = int(imageheight * printres + 0.5)
            # Set up to draw to the QPrinter
            painter = QPainter(printer)
            if not transparent:
                # Draw a rectangle filling the entire scene
                # with the last clearing color.
                # Only draw if not completely transparent
                if (self.__lastclearcolor.getRgb())[3] > 0:
                    painter.fillRect(QRectF(0, 0, printwidth, printheight),
                                     self.__lastclearcolor)
            # Scaling printfactor for the scene to the saved image
            widthscalefactor = imagewidth * self.physicalDpiX() / float(self.__scenewidth)
            # Check if there are annotations to add
            if annopicture:
                # Scale the scene now for the annotations
                painter.scale(widthscalefactor, widthscalefactor)
                # factor that makes it work after scaling (12.5 = 1200 / 96)
                printfactor = printres / self.physicalDpiX()
                # Draw a solid white rectangle with black outline for the annotations
                painter.setBrush(QBrush(Qt.white, Qt.SolidPattern))
                painter.setPen(QPen(QBrush(Qt.black, Qt.SolidPattern),
                                    2.0 * printfactor, Qt.SolidLine, Qt.SquareCap, Qt.MiterJoin))
                painter.drawRect(QRectF(1.0 * printfactor, 1.0 * printfactor,
                    (self.__scenewidth - 2.0) * printfactor,
                    ((annosize.height() + 2.0 * self.__addedannomargin) - 2.0) * printfactor))
                # And add the annotations within this box
                painter.drawPicture(QPointF(self.__addedannomargin * printfactor,
                                            self.__addedannomargin * printfactor),
                                    annopicture)
                # Draw the scene to the printer - scaling already in effect
                self.paintScene(painter, 0, 0.0,
                        (annosize.height() + 2.0 * self.__addedannomargin) * printfactor,
                        1.0, "Saving", False)
            else:
                # No annotations so just do the normal drawing
                self.paintScene(painter, 0, 0.0, 0.0,
                                widthscalefactor, "Saving", False)
            if (self.__wmarkFilename is not None):
                # Initialize watermark objects
                wmkpic = QPixmap(self.__wmarkFilename)
                wmkpt = QPointF()
                wmkpt.setX(self.__xloc)
                wmkpt.setY(self.__yloc)
                # set watermark image display opacity
                painter.setOpacity(self.__opacity)
                # set image scale
                painter.scale(self.__scalefrac, self.__scalefrac)
                # paint watermark image at specified location
                painter.setRenderHint(QPainter.Antialiasing)
                painter.drawPixmap(wmkpt, wmkpic)
            painter.end()
        elif myformat == 'svg':
            generator = QSvgGenerator()
            generator.setFileName(myfilename)
            if vectsize:
                imagewidth = int(vectsize.width() * self.physicalDpiX() + 0.5)
                imageheight = int(vectsize.height() * self.physicalDpiY() + 0.5)
            else:
                imagewidth = int(self.__scenewidth * self.__scalefactor + 0.5)
                imageheight = int(self.__sceneheight * self.__scalefactor + 0.5)
            # Add in any height needed for the annotations
            if annopicture:
                annoheight = (annosize.height() + 2 * self.__addedannomargin) * \
                             imageheight / self.__sceneheight
                imageheight += annoheight
            # Set the image size
            generator.setResolution(
                    int(0.5 * (self.physicalDpiX() + self.physicalDpiY()) + 0.5) )
            generator.setSize( QSize(imagewidth, imageheight) )
            generator.setViewBox( QRect(0, 0, imagewidth, imageheight) )
            # paint the scene to this QSvgGenerator
            painter = QPainter(generator)
            if not transparent:
                # Draw a rectangle filling the entire scene
                # with the last clearing color.
                # Only draw if not completely transparent
                if (self.__lastclearcolor.getRgb())[3] > 0:
                    painter.fillRect( QRectF(0, 0, imagewidth, imageheight),
                                      self.__lastclearcolor )
            # Scaling printfactor for the scene to the saved image
            widthscalefactor = imagewidth / float(self.__scenewidth)
            # Check if there are annotations to add
            if annopicture:
                # Scale the scene now for the annotations
                painter.scale(widthscalefactor, widthscalefactor)
                # Draw a solid white rectangle with black outline for the annotations
                painter.setBrush(QBrush(Qt.white, Qt.SolidPattern))
                painter.setPen(QPen(QBrush(Qt.black, Qt.SolidPattern),
                                    2.0, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin))
                painter.drawRect(QRectF(1.0, 1.0,
                            self.__scenewidth - 2.0,
                            annosize.height() + 2.0 * self.__addedannomargin - 2.0))
                # And add the annotations within this box
                painter.drawPicture(QPointF(self.__addedannomargin,self.__addedannomargin),
                                    annopicture)
                # Draw the scene to the printer - scaling already in effect
                self.paintScene(painter, 0,
                                0.0, annosize.height() + 2.0 * self.__addedannomargin,
                                1.0, "Saving", False)
            else:
                # No annotations so just do the normal drawing
                self.paintScene(painter, 0, 0.0, 0.0,
                                widthscalefactor, "Saving", False)
            if (self.__wmarkFilename is not None):
                # Initialize watermark objects
                wmkpic = QPixmap(self.__wmarkFilename)
                wmkpt = QPointF()
                wmkpt.setX(self.__xloc)
                wmkpt.setY(self.__yloc)
                # set watermark image display opacity
                painter.setOpacity(self.__opacity)
                # set image scale
                painter.scale(self.__scalefrac, self.__scalefrac)
                # paint watermark image at specified location
                painter.setRenderHint(QPainter.Antialiasing)
                painter.drawPixmap(wmkpt, wmkpic)
            painter.end()
        else:
            if rastsize:
                imagewidth = int(rastsize.width() + 0.5)
                imageheight = int(rastsize.height() + 0.5)
            else:
                imagewidth = int(self.__scenewidth * self.__scalefactor + 0.5)
                imageheight = int(self.__sceneheight * self.__scalefactor + 0.5)
            # Add in any height needed for the annotations
            if annopicture:
                annoheight = (annosize.height() + 2 * self.__addedannomargin) * \
                             imageheight / self.__sceneheight
                imageheight += annoheight
            # Create the image
            image = QImage( QSize(imagewidth, imageheight),
                            QImage.Format_ARGB32_Premultiplied )
            # Indicate the recommended displayed size of PNG images
            image.setDotsPerMeterX(self.physicalDpiX() / 0.0254)
            image.setDotsPerMeterY(self.physicalDpiY() / 0.0254)
            # Initialize the image
            # Note that completely transparent gives black for formats not supporting
            # the alpha channel (JPEG) whereas ARGB32 with 0x00FFFFFF gives white
            if not transparent:
                # Clear the image with self.__lastclearcolor
                fillint = self.__helper.computeARGB32PreMultInt(self.__lastclearcolor)
            else:
                fillint = 0
            image.fill(fillint)
            # paint the scene to this QImage
            painter = QPainter(image)
            # Scaling printfactor for the scene to the saved image
            widthscalefactor = imagewidth / float(self.__scenewidth)
            # Check if there are annotations to add
            if annopicture:
                # Scale the scene now for the annotations
                painter.scale(widthscalefactor, widthscalefactor)
                # Draw a solid white rectangle with black outline for the annotations
                painter.setBrush(QBrush(Qt.white, Qt.SolidPattern))
                painter.setPen(QPen(QBrush(Qt.black, Qt.SolidPattern),
                                    2.0, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin))
                painter.drawRect(QRectF(1.0, 1.0,
                            self.__scenewidth - 2.0,
                            annosize.height() + 2.0 * self.__addedannomargin - 2.0))
                # And add the annotations within this box
                painter.drawPicture(QPointF(self.__addedannomargin,self.__addedannomargin),
                                    annopicture)
                # Draw the scene to the printer - scaling already in effect
                self.paintScene(painter, 0,
                                0.0, annosize.height() + 2.0 * self.__addedannomargin,
                                1.0, "Saving", False)
            else:
                # No annotations so just do the normal drawing
                self.paintScene(painter, 0, 0.0, 0.0,
                                widthscalefactor, "Saving", False)
            if (self.__wmarkFilename is not None):
                # Initialize watermark objects
                wmkpic = QPixmap(self.__wmarkFilename)
                wmkpt = QPointF()
                wmkpt.setX(self.__xloc)
                wmkpt.setY(self.__yloc)
                # set watermark image display opacity
                painter.setOpacity(self.__opacity)
                # set image scale
                painter.scale(self.__scalefrac, self.__scalefrac)
                # paint watermark image at specified location
                painter.setRenderHint(QPainter.Antialiasing)
                painter.drawPixmap(wmkpt, wmkpic)
            painter.end()
            # save the image to file
            if not image.save(myfilename, myformat):
                raise ValueError("Unable to save the plot as " + myfilename)

    def checkCommandPipe(self):
        '''
        Get and perform commands waiting in the pipe.
        Stop when no more commands or if more than 50
        milliseconds have passed.
        '''
        try:
            if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 3):
                starttime = time.process_time()
            else:
                starttime = time.clock()
            # Wait up to 2 milliseconds waiting for a command.
            # This prevents unchecked spinning when there is
            # nothing to do (Qt immediately calling this method
            # again only for this method to immediately return).
            while self.__cmndpipe.poll(0.002):
                cmnd = self.__cmndpipe.recv()
                self.processCommand(cmnd)
                # Continue to try to process commands until
                # more than 50 milliseconds have passed.
                # This reduces Qt overhead when there are lots
                # of commands waiting in the queue.
                if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 3):
                    elapsed = time.process_time() - starttime
                else:
                    elapsed = time.clock() - starttime
                if elapsed > 0.050:
                    break
        except EOFError:
            # Assume PyFerret has shut down
            self.exitViewer()
        except Exception:
            # Some problem, but presumably still functional
            (exctype, excval) = sys.exc_info()[:2]
            try:
                if excval:
                    self.__rspdpipe.send("**ERROR %s: %s" % (str(exctype), str(excval)))
                else:
                    self.__rspdpipe.send("**ERROR %s" % str(exctype))
            except Exception:
                pass

    def processCommand(self, cmnd):
        '''
        Examine the action of cmnd and call the appropriate
        method to deal with this command.  Raises a ValueError
        if the "action" key is missing.
        '''
        try:
            cmndact = cmnd["action"]
        except KeyError:
            raise ValueError("Unknown command '%s'" % str(cmnd))

        if cmndact == "clear":
            try:
                bkgcolor = self.__helper.getColorFromCmnd(cmnd)
            except KeyError:
                bkgcolor = None
            self.clearScene(bkgcolor)
        elif cmndact == "exit":
            self.exitViewer()
        elif cmndact == "hide":
            self.showMinimized()
        elif cmndact == "screenInfo":
            scrnrect = QApplication.desktop().availableGeometry()
            info = ( self.physicalDpiX(), self.physicalDpiY(),
                     scrnrect.width(), scrnrect.height() )
            self.__rspdpipe.send(info)
        elif cmndact == "antialias":
            self.__antialias = bool(cmnd.get("antialias", True))
        elif cmndact == "update":
            self.updateScene()
        elif cmndact == "redraw":
            try:
                bkgcolor = self.__helper.getColorFromCmnd(cmnd)
            except KeyError:
                bkgcolor = None
            self.redrawScene(bkgcolor)
        elif cmndact == "rescale":
            newscale = float(cmnd["factor"])
            self.scaleScene(newscale, True)
        elif cmndact == "resize":
            mysize = self.__helper.getSizeFromCmnd(cmnd)
            self.resizeScene(mysize.width(), mysize.height())
        elif cmndact == "save":
            filename = cmnd["filename"]
            fileformat = cmnd.get("fileformat", None)
            transparent = cmnd.get("transparent", False)
            vectsize = self.__helper.getSizeFromCmnd(cmnd["vectsize"])
            rastsize = self.__helper.getSizeFromCmnd(cmnd["rastsize"])
            try:
                myannotations = cmnd["annotations"]
            except KeyError:
                myannotations = None
            self.saveSceneToFile(filename, fileformat, transparent,
                                 vectsize, rastsize, myannotations)
        elif cmndact == "setWidthFactor":
            newfactor = float(cmnd.get("factor", -1.0))
            if newfactor <= 0.0:
                raise ValueError("Invalid width factor")
            self.setWidthScalingFactor(newfactor)
        elif cmndact == "setTitle":
            self.setWindowTitle(cmnd["title"])
        elif cmndact == "imgname":
            value = cmnd.get("name", None)
            if value:
                self.__lastfilename = value
            value = cmnd.get("format", None)
            if value:
                self.__lastformat = value.lower()
        elif cmndact == "show":
            if not self.isVisible():
                self.show()
        elif cmndact == "noalpha":
            self.__noalpha = True
        elif cmndact == "beginView":
            self.beginView(cmnd)
        elif cmndact == "clipView":
            self.clipView(cmnd)
        elif cmndact == "endView":
            self.endView(True)
        elif cmndact == "beginSegment":
            self.beginSegment(cmnd["segid"])
        elif cmndact == "endSegment":
            self.endSegment(True)
        elif cmndact == "deleteSegment":
            self.deleteSegment(cmnd["segid"])
        elif cmndact == "createSymbol":
            # Define this symbol in self.__symbolpaths
            sympath = self.__helper.getSymbolFromCmnd(cmnd)
            # The name is now all that is needed to use this symbol
            self.__rspdpipe.send(cmnd['name'])
        elif cmndact == "drawMultiline":
            self.drawMultiline(cmnd)
        elif cmndact == "drawPoints":
            self.drawPoints(cmnd)
        elif cmndact == "drawPolygon":
            self.drawPolygon(cmnd)
        elif cmndact == "drawRectangle":
            self.drawRectangle(cmnd)
        elif cmndact == "textSize":
            info = self.getSimpleTextSize(cmnd)
            self.__rspdpipe.send(info)
        elif cmndact == "drawText":
            self.drawSimpleText(cmnd)
        elif cmndact == "setWaterMark":
            # self.setWaterMark(cmnd)
            self.setWaterMark(cmnd['filename'], None, cmnd['xloc'], cmnd['yloc'], cmnd['scalefrac'], cmnd['opacity'])
        else:
            raise ValueError("Unknown command action %s" % str(cmndact))

    def beginView(self, cmnd):
        '''
        Setup a new viewport and window for drawing on a portion
        (possibly all) of the scene.  Recognized keys from cmnd
        are:
            "viewfracs": a dictionary of sides positions (see
                    CmndHelperPQ.getSidesFromCmnd) giving the
                    fractions [0.0, 1.0] of the way through the
                    scene for the sides of the new View.
            "clip": clip to the new View? (default: True)

        Note that the view fraction values are based on (0,0) being the
        top left corner and (1,1) being the bottom right corner.  Thus,
        left < right and top < bottom.

        Raises a KeyError if the "viewfracs" key is not given.
        '''
        # Get the view rectangle in fractions of the full scene
        fracsides = self.__helper.getSidesFromCmnd(cmnd["viewfracs"])
        # Should graphics be clipped to this view?
        try:
            clipit = cmnd["clip"]
        except KeyError:
            clipit = True
        self.beginViewFromSides(fracsides, clipit)

    def beginViewFromSides(self, fracsides, clipit):
        '''
        Setup a new viewport and window for drawing on a portion
        (possibly all) of the scene.  The view in fractions of
        the full scene are given in fracsides.  Sets the clipping
        rectangle to this view.  If clipit is True, graphics
        will be clipped to this view.
        '''
        # If a view is still active, automatically end it
        if self.__activepainter:
            self.endView(True)
        # Get the location for the new view in terms of scene pixels.
        width = float(self.__scenewidth)
        height = float(self.__sceneheight)
        leftpixel = fracsides.left() * width
        rightpixel = fracsides.right() * width
        bottompixel = fracsides.bottom() * height
        toppixel = fracsides.top() * height
        # perform the checks after turning into units of pixels
        # to make sure the values are significantly different
        if (0.0 > leftpixel) or (leftpixel >= rightpixel) or (rightpixel > width):
            raise ValueError("Invalid left, right view fractions: " \
                             "left in pixels = %s, right in pixels = %s" \
                             % (str(leftpixel), str(rightpixel)) )
        if (0.0 > toppixel) or (toppixel >= bottompixel) or (bottompixel > height):
            raise ValueError("Invalid bottom, top view fractions: " \
                             "top in pixels = %s, bottom in pixels = %s" \
                             % (str(toppixel), str(bottompixel)) )
        # Create the view rectangle in device coordinates
        vrectf = QRectF(leftpixel, toppixel,
                       rightpixel - leftpixel, bottompixel - toppixel)
        # Create the new picture and painter
        self.__activepicture = QPicture()
        self.__activepainter = QPainter(self.__activepicture)
        # Set the clip rectangle to that of the view; this also activates clipping
        self.__activepainter.setClipRect(vrectf, Qt.ReplaceClip)
        # Disable clipping if not desired at this time
        if not clipit:
            self.__activepainter.setClipping(False)
        # Note that __activepainter has to end before __activepicture will
        # draw anything.  So no need to add it to __viewpics until then.
        self.__drawcount = 0
        # Save the current view sides and clipit setting for recreating the view.
        # Just save the original objects (assume calling functions do not keep them)
        self.__fracsides = fracsides
        self.__clipit = clipit

    def clipView(self, cmnd):
        '''
        If cmnd["clip"] is True, activates clipping to the
        current view rectangle.  If cmnd["clip"] is False,
        disable clipping in this view.

        Raises a KeyError if the "clip" key is not given.
        '''
        if cmnd["clip"]:
            self.__activepainter.setClipping(True)
            self.__clipit = True
        else:
            self.__activepainter.setClipping(False)
            self.__clipit = False

    def endView(self, update):
        '''
        Ends the current view and appends it to the list of pictures
        drawn in the scene.  If update is True, the displayed scene
        is updated.
        '''
        self.__activepainter.end()
        self.__activepainter = None
        # Only save the active picture if it contains something
        if self.__drawcount > 0:
            self.__viewpics.append((self.__activepicture, self.__segid))
            self.__drawcount = 0
        if update:
            # update the scene
            self.drawLastPictures(False)
        self.__activepicture = None

    def beginSegment(self, segid):
        '''
        Associates a segment ID with the current empty view
        (picture) and all future views until endSegment is
        called.  If the current view is not empty, the current
        view is ended and a new view started.  If there is not
        a active view, the segment ID is just saved for the
        next active view.
        '''
        if self.__activepainter and (self.__drawcount > 0):
            self.endView(True)
            self.beginViewFromSides(self.__fracsides, self.__clipit)
        self.__segid = segid

    def endSegment(self, update):
        '''
        Ends the current active view and starts a new view.
        Removes the current segment ID associated with views.
        '''
        if self.__activepainter and (self.__drawcount > 0):
            self.endView(update)
            self.beginViewFromSides(self.__fracsides, self.__clipit)
        if update:
            self.drawLastPictures(False)
        self.__segid = None

    def deleteSegment(self, segid):
        '''
        Removes all pictures associated with the given segment ID
        '''
        # if deleting the current segment, end the current segment
        if segid == self.__segid:
            self.endSegment(False)
        # Go through all the pictures, determining which to save
        newpicts = [ ]
        for (viewpic, vsegid) in self.__viewpics:
            if vsegid != segid:
                newpicts.append((viewpic, vsegid))
            else:
                # picture was deleted, so will need to
                # regenerate the scene from the beginning
                self.__clearpixmap = True
                self.__lastpicdrawn = 0
        self.__viewpics[:] = newpicts
        # Do NOT update since there may be more segments to be deleted
        # Rely on the receiving an update or redraw command at the end

    def updateScene(self):
        '''
        Updates the displayed graphics to include all drawn elements.
        '''
        # If there is an active picture containing something,
        # end the view, thus adding and display this picture,
        # then restart the view.
        if self.__drawcount > 0:
            self.endView(True)
            self.beginViewFromSides(self.__fracsides, self.__clipit)
        self.drawLastPictures(False)

    def drawMultiline(self, cmnd):
        '''
        Draws a collection of connected line segments.

        Recognized keys from cmnd:
            "points": consecutive endpoints of the connected line
                    segments as a list of (x, y) coordinates
            "pen": dictionary describing the pen used to draw the
                    segments (see CmndHelperPQ.getPenFromCmnd)

        The coordinates are device coordinates from the upper left corner.

        Raises:
            KeyError if the "points" or "pen" key is not given
            ValueError if there are fewer than two endpoints given
        '''
        ptcoords = cmnd["points"]
        if len(ptcoords) < 2:
            raise ValueError("fewer that two endpoints given")
        endpts = QPolygonF( [ QPointF(xypair[0], xypair[1]) \
                                  for xypair in ptcoords ] )
        mypen = self.__helper.getPenFromCmnd(cmnd["pen"])
        self.__activepainter.setRenderHint(QPainter.Antialiasing,
                                           self.__antialias)
        self.__activepainter.setBrush(Qt.NoBrush)
        self.__activepainter.setPen(mypen)
        self.__activepainter.drawPolyline(endpts)
        self.__drawcount += 1
        # Limit the number of drawing commands per picture
        if self.__drawcount >= self.__maxdraws:
            self.updateScene()

    def drawPoints(self, cmnd):
        '''
        Draws a collection of discrete points using a single symbol
        for each point.

        Recognized keys from cmnd:
            "points": point centers as a list of (x,y) coordinates
            "symbol": symbol to use (see CmndHelperPQ.getSymbolFromCmnd)
            "size": size of the symbol in points (1/72 inches) before
                    scaling by the width scaling factor
            "color": color name or 24-bit RGB integer value (eg, 0xFF0088)
            "alpha": alpha value from 0 (transparent) to 255 (opaque)
            "highlight": dictionary of "color" and "alpha" (as above)
                     for filled symbol outline color; not outlined if omitted

        The coordinates are device coordinates from the upper left corner.

        Raises a KeyError if the "symbol", "points", or "size" key
        is not given.
        '''
        ptcoords = cmnd["points"]
        ptsize = cmnd["size"]
        try:
            highlight = self.__helper.getColorFromCmnd(cmnd["highlight"])
        except KeyError:
            highlight = None
        sympath = self.__helper.getSymbolFromCmnd(cmnd["symbol"])
        self.__activepainter.setRenderHint(QPainter.Antialiasing,
                                           self.__antialias)
        try:
            mycolor = self.__helper.getColorFromCmnd(cmnd)
            mybrush = QBrush(mycolor, Qt.SolidPattern)
        except KeyError:
            mybrush = QBrush(Qt.SolidPattern)
        if sympath.isFilled():
            self.__activepainter.setBrush(mybrush)
            if highlight:
                # highlighted filled plot - pen width is 4% of the width of the symbol
                mybrush = QBrush(highlight, Qt.SolidPattern)
                mypen = QPen(mybrush, 4.0, Qt.SolidLine, Qt.FlatCap, Qt.MiterJoin)
                self.__activepainter.setPen(mypen)
            else:
                # filled plot without highlight - no pen, only brush
                self.__activepainter.setPen(Qt.NoPen)
        else:
            # stroked path - no brush, pen width is 8% of the width of the symbol, highlight is ignored
            self.__activepainter.setBrush(Qt.NoBrush)
            mypen = QPen(mybrush, 8.0, Qt.SolidLine, Qt.FlatCap, Qt.MiterJoin)
            self.__activepainter.setPen(mypen)
        # typical symbols are 100x100 pixels
        scalefactor = ptsize * self.widthScalingFactor() / 100.0
        if self.__maxsymbolwidth < 100.0 * scalefactor:
            self.__maxsymbolwidth = 100.0 * scalefactor
        if self.__maxsymbolheight < 100.0 * scalefactor:
            self.__maxsymbolheight = 100.0 * scalefactor
        for xyval in ptcoords:
            # save so the translation and scale are not permanent
            self.__activepainter.save()
            try:
                self.__activepainter.translate( QPointF(xyval[0], xyval[1]) )
                self.__activepainter.scale(scalefactor, scalefactor)
                self.__activepainter.drawPath(sympath.painterPath())
            finally:
                self.__activepainter.restore()
        self.__drawcount += len(ptcoords)
        # Limit the number of drawing commands per picture
        if self.__drawcount >= self.__maxdraws:
            self.updateScene()

    def drawPolygon(self, cmnd):
        '''
        Draws a polygon item to the viewer.

        Recognized keys from cmnd:
            "points": the vertices of the polygon as a list of (x,y)
                    coordinates
            "fill": dictionary describing the brush used to fill the
                    polygon; see CmndHelperPQ.getBrushFromCmnd
                    If not given, the polygon will not be filled.
            "outline": dictionary describing the pen used to outline
                    the polygon; see CmndHelperPQ.getPenFromCmnd
                    If not given, the border will be drawn with a
                    cosmetic pen identical to the brush used to fill
                    the polygon.

        The coordinates are device coordinates from the upper left corner.

        Raises a KeyError if the "points" key is not given.
        '''
        mypoints = cmnd["points"]
        mypolygon = QPolygonF( [ QPointF(xypair[0], xypair[1]) \
                                     for xypair in mypoints ] )
        self.__activepainter.setRenderHint(QPainter.Antialiasing,
                                           False)
        try:
            mybrush = self.__helper.getBrushFromCmnd(cmnd["fill"])
        except KeyError:
            mybrush = Qt.NoBrush
        try:
            mypen = self.__helper.getPenFromCmnd(cmnd["outline"])
        except KeyError:
            if ( mybrush == Qt.NoBrush ):
                raise ValueError('drawPolygon called without a Brush or Pen')
            # Use a "cosmetic" Pen matching the brush
            # mypen = QPen(mybrush, 0.0, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin)
            mypen = Qt.NoPen
        self.__activepainter.setBrush(mybrush)
        self.__activepainter.setPen(mypen)
        self.__activepainter.drawPolygon(mypolygon)
        self.__drawcount += 1
        # Limit the number of drawing commands per picture
        if self.__drawcount >= self.__maxdraws:
            self.updateScene()

    def drawRectangle(self, cmnd):
        '''
        Draws a rectangle in the current view using the information
        in the dictionary cmnd.

        Recognized keys from cmnd:
            "left": x-coordinate of left edge of the rectangle
            "bottom": y-coordinate of the bottom edge of the rectangle
            "right": x-coordinate of the right edge of the rectangle
            "top": y-coordinate of the top edge of the rectangle
            "fill": dictionary describing the brush used to fill the
                    rectangle; see CmndHelperPQ.getBrushFromCmnd
                    If not given, the rectangle will not be filled.
            "outline": dictionary describing the pen used to outline
                    the rectangle; see CmndHelperPQ.getPenFromCmnd
                    If not given, the border will be drawn with a
                    cosmetic pen identical to the brush used to fill
                    the rectangle.

        The coordinates are device coordinates from the upper left corner.

        Raises a ValueError if the width or height of the rectangle
        is not positive.
        '''
        # get the left, bottom, right, and top values
        # any keys not given get a zero value
        sides = self.__helper.getSidesFromCmnd(cmnd)
        width = sides.right() - sides.left()
        if width <= 0.0:
            raise ValueError("width of the rectangle (%s) in not positive" % str(width))
        height = sides.bottom() - sides.top()
        if height <= 0.0:
            raise ValueError("height of the rectangle (%s) in not positive" % str(height))
        myrect = QRectF(sides.left(), sides.top(), width, height)
        self.__activepainter.setRenderHint(QPainter.Antialiasing,
                                           False)
        try:
            mybrush = self.__helper.getBrushFromCmnd(cmnd["fill"])
        except KeyError:
            mybrush = Qt.NoBrush
        try:
            mypen = self.__helper.getPenFromCmnd(cmnd["outline"])
        except KeyError:
            if ( mybrush == Qt.NoBrush ):
                raise ValueError('drawPolygon called without a Brush or Pen')
            # Use a "cosmetic" Pen matching the brush
            # mypen = QPen(mybrush, 1.0, Qt.SolidLine, Qt.SquareCap, Qt.BevelJoin)
            mypen = Qt.NoPen
        self.__activepainter.setBrush(mybrush)
        self.__activepainter.setPen(mypen)
        self.__activepainter.drawRect(myrect)
        self.__drawcount += 1
        # Limit the number of drawing commands per picture
        if self.__drawcount >= self.__maxdraws:
            self.updateScene()

    def getSimpleTextSize(self, cmnd):
        '''
        Returns the pair (width, height) for given text when drawn.
        Raises a KeyError if the "text" key is not given.

        The width value is the width for the text that can be used
        for positioning the next text item to draw.  The height
        value is the ascent plus decent for the font and does not
        depend of the text.  The bounding rectangle for the actual
        drawn text may exceed this (width, height) if,
        e.g., italic or unusual characters.

        Recognized keys from cmnd:
            "text": string to displayed
            "font": dictionary describing the font to use;  see
                    CmndHelperPQ.getFontFromCmnd.  If not given
                    the default font for this viewer is used.
        '''
        try:
            myfont = self.__helper.getFontFromCmnd(cmnd["font"])
        except KeyError:
            myfont = self.__activepainter.font()
        myfontmetrics = QFontMetricsF(myfont)
        mytext = cmnd["text"]
        if PYTHONQT_VERSION == 'PyQt4':
            mytext = QString.fromUtf8(mytext)
        width = myfontmetrics.width(mytext)
        height = myfontmetrics.height()
        return (width, height)

    def drawSimpleText(self, cmnd):
        '''
        Draws a "simple" text item in the current view.
        Raises a KeyError if the "text" or "location" key is not given.

        Recognized keys from cmnd:
            "text": null-terminated UTF-8 encoded string to be displayed
            "font": dictionary describing the font to use;  see
                    CmndHelperPQ.getFontFromCmnd.  If not given
                    the default font for this viewer is used.
            "fill": dictionary describing the pen used to draw the
                    text; see CmndHelperPQ.getPenFromCmnd.
                    If not given, the default pen for this viewer
                    is used.
            "rotate": clockwise rotation of the text in degrees
            "location": (x,y) location for the baseline of the
                    start of text.  The coordinates are device
                    coordinates from the upper left corner.
        '''
        mytext = cmnd["text"]
        startpt = cmnd["location"]
        self.__activepainter.setRenderHint(QPainter.Antialiasing,
                                           self.__antialias)
        self.__activepainter.setBrush(Qt.NoBrush)
        try:
            mypen = self.__helper.getPenFromCmnd(cmnd["fill"])
            self.__activepainter.setPen(mypen)
        except KeyError:
            pass
        # save so the font, translation, and rotation are not permanent
        self.__activepainter.save()
        try:
            try:
                myfont = self.__helper.getFontFromCmnd(cmnd["font"])
                self.__activepainter.setFont(myfont)
            except KeyError:
                pass
            # Move the coordinate system so the origin is at the start
            # of the text so that rotation is about this point
            self.__activepainter.translate(startpt[0], startpt[1])
            try:
                rotdeg = cmnd["rotate"]
                self.__activepainter.rotate(rotdeg)
            except KeyError:
                pass

            if PYTHONQT_VERSION == 'PyQt4':
                mytext = QString.fromUtf8(mytext)
            self.__activepainter.drawText(0, 0, mytext)
            self.__drawcount += 1
        finally:
            # return the painter to the default state
            self.__activepainter.restore()
        # Limit the number of drawing commands per picture
        if self.__drawcount >= self.__maxdraws:
            self.updateScene()

    def setWidthScalingFactor(self, factor):
        '''
        Assign the scaling factor for line widths and symbol sizes
        to convert from points (1/72 inches) to pixels, and to apply
        any additional width scaling specified by factor.
        '''
        self.__widthfactor  = (self.physicalDpiX() + self.physicalDpiY()) / 144.0
        self.__widthfactor *= factor

    def widthScalingFactor(self):
        '''
        Return the scaling factor for line widths and symbol sizes
        to convert from points (1/72 inches) to pixels, and to apply
        any additional width scaling specified by setWidthFactor.
        '''
        return self.__widthfactor

    def setWaterMark(self, filename, len_filename, xloc, yloc, scalefrac, opacity):
        '''
        Overlay watermark from contents of filename.

        Recognized keys from cmnd:
            "filename":  water mark image file
            "xloc":      horizontal position of upper left corner of watermark image
            "yloc":      vertical position of upper left corner of watermark image
            "scalefrac": multiple of original image size to display plot as
            "opacity":   image visibility in range [0.0, 1.0] where 0->invisible, 1->opaque
        '''
        self.__wmarkFilename = str(filename)
        self.__xloc = xloc
        self.__yloc = yloc
        self.__scalefrac = scalefrac
        self.__opacity = opacity

class PipedViewerPQProcess(Process):
    '''
    A Process specifically tailored for creating a PipedViewerPQ.
    '''
    def __init__(self, cmndpipe, rspdpipe):
        '''
        Create a Process that will produce a PipedViewerPQ
        attached to the given Pipes when run.
        '''
        super(PipedViewerPQProcess,self).__init__(group=None, target=None, name='PipedViewerPQ')
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        self.__app = None
        self.__viewer = None

    def run(self):
        '''
        Create a PipedViewerPQ that is attached
        to the Pipe of this instance.
        '''
        self.__app = QApplication(["PipedViewerPQ"])
        self.__viewer = PipedViewerPQ(self.__cmndpipe, self.__rspdpipe)
        myresult = self.__app.exec_()
        sys.exit(myresult)


#
# The following are for testing this (and the cmndhelperpq) modules
#

class _CommandSubmitterPQ(QDialog):
    '''
    Testing dialog for controlling the addition of commands to a pipe.
    Used for testing PipedViewerPQ in the same process as the viewer.
    '''
    def __init__(self, parent, cmndpipe, rspdpipe, cmndlist):
        '''
        Create a QDialog with a single QPushButton for controlling
        the submission of commands from cmndlist to cmndpipe.
        '''
        super(_CommandSubmitterPQ,self).__init__(parent)
        self.__cmndlist = cmndlist
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        self.__nextcmnd = 0
        self.__button = QPushButton("Submit next command", self)
        self.__button.pressed.connect(self.submitNextCommand)
        self.show()

    def submitNextCommand(self):
        '''
        Submit the next command from the command list to the command pipe,
        or shutdown if there are no more commands to submit.
        '''
        try:
            print("Command: %s" % str(self.__cmndlist[self.__nextcmnd]))
            self.__cmndpipe.send(self.__cmndlist[self.__nextcmnd])
            self.__nextcmnd += 1
            while self.__rspdpipe.poll(0.1):
                print("Response: %s" % str(self.__rspdpipe.recv()))
        except IndexError:
            self.__rspdpipe.close()
            self.__cmndpipe.close()
            self.close()


def _test_pipedviewerpq():
    # vertices of a pentagon (roughly) centered in a 1000 x 1000 square
    pentagonpts = ( (504.5, 100.0), (100.0, 393.9),
                    (254.5, 869.4), (754.5, 869.4),
                    (909.0, 393.9),  )

    # create the list of commands to submit
    drawcmnds = []
    drawcmnds.append( { "action":"setTitle", "title":"Tester" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"clear", "color":"black"} )
    drawcmnds.append( { "action":"screenInfo"} )
    drawcmnds.append( { "action":"antialias", "antialias":True } )
    drawcmnds.append( { "action":"resize",
                        "width":500,
                        "height":500 } )
    drawcmnds.append( { "action":"beginView",
                        "viewfracs":{"left":0.0, "right":0.5,
                                     "top":0.5, "bottom":1.0},
                        "clip":True } )
    drawcmnds.append( { "action":"drawRectangle",
                        "left": 5, "right":245,
                        "top":245, "bottom":495,
                        "fill":{"color":"green", "alpha":128} } )
    mypentapts = [ (.25 * ptx, .25 * pty + 250) for (ptx, pty) in pentagonpts ]
    drawcmnds.append( { "action":"drawPolygon",
                        "points":mypentapts,
                        "fill":{"color":"blue"},
                        "outline":{"color":"black",
                                   "width": 5,
                                   "style":"solid",
                                   "capstyle":"round",
                                   "joinstyle":"round" } } )
    drawcmnds.append( { "action":"beginSegment",
                        "segid":"text" } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=480",
                        "font":{"family":"Times", "size":16},
                        "fill":{"color":"red"},
                        "location":(50,480) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=430",
                        "font":{"family":"Times", "size":16},
                        "fill":{"color":"red"},
                        "location":(50,430) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=380",
                        "font":{"family":"Times", "size":16},
                        "fill":{"color":"red"},
                        "location":(50,380) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=330",
                        "font":{"family":"Times", "size":16},
                        "fill":{"color":"red"},
                        "location":(50,330) } )
    drawcmnds.append( { "action":"textSize",
                        "text":"This is a some line of text",
                        "font":{"family":"Times", "size":16} } )
    drawcmnds.append( { "action":"endSegment" } )
    drawcmnds.append( { "action":"endView" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"createSymbol",
                        "name": "uptrifill",
                        "pts": ( (-40.0, -30.0), (0.0, 40.0), (40.0, -30.0), (-40.0, -30.0), ),
                        "fill": True } )
    drawcmnds.append( { "action":"createSymbol",
                        "name": "bararrow",
                        "pts": ( (-50,50), (-10,10),
                                 (-999, -999),
                                 (50,0), (50,50), (0,50),
                                 (-999, -999),
                                 (0,-10), (20,-30), (10,-30), (10,-50), (-10,-50), (-10,-30), (-20,-30), (0,-10), ),
                        "fill": False } )
    drawcmnds.append( { "action":"beginView",
                        "viewfracs":{"left":0.0, "right":1.0,
                                     "top":0.0, "bottom":1.0},
                        "clip":True } )
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (100,  50),
                                   (100, 150),
                                   (100, 250),
                                   (100, 350),
                                   (100, 450) ),
                        "symbol":"dot",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (150,  50),
                                   (150, 150),
                                   (150, 250),
                                   (150, 350),
                                   (150, 450) ),
                        "symbol":"circle",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (200,  50),
                                   (200, 150),
                                   (200, 250),
                                   (200, 350),
                                   (200, 450) ),
                        "symbol":"dotplus",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (250,  50),
                                   (250, 150),
                                   (250, 250),
                                   (250, 350),
                                   (250, 450) ),
                        "symbol":"circplus",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (300,  50),
                                   (300, 150),
                                   (300, 250),
                                   (300, 350),
                                   (300, 450) ),
                        "symbol":"dotex",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (350,  50),
                                   (350, 150),
                                   (350, 250),
                                   (350, 350),
                                   (350, 450) ),
                        "symbol":"circex",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (400,  50),
                                   (400, 150),
                                   (400, 250),
                                   (400, 350),
                                   (400, 450) ),
                        "symbol":"uptrifill",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (450,  50),
                                   (450, 150),
                                   (450, 250),
                                   (450, 350),
                                   (450, 450) ),
                        "symbol":"bararrow",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawMultiline",
                        "points":( (350,  50),
                                   (200, 150),
                                   (400, 250),
                                   (300, 350),
                                   (150, 250),
                                   (100, 450) ),
                        "pen": {"color":"white",
                                "width":3,
                                "style":"dash",
                                "capstyle":"round",
                                "joinstyle":"round"} } )
    drawcmnds.append( { "action":"endView" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"deleteSegment",
                        "segid":"text" } )
    drawcmnds.append( { "action":"update" } )
    drawcmnds.append( { "action":"show" } )
    annotations = ( "The 1<sup>st</sup> CO<sub>2</sub> annotations line",
                    "Another line with <i>lengthy</i> details that go on and on " + \
                    "and on and should wrap to a 2<sup>nd</sup> annotation line",
                    "<b>Final</b> annotation line" )
    drawcmnds.append( { "action":"save",
                        "filename":"test.pdf",
                        "vectsize":{"width":7.0, "height":7.0},
                        "rastsize":{"width":750, "height":750},
                        "annotations":annotations } )
    drawcmnds.append( { "action":"save",
                        "filename":"test.png",
                        "vectsize":{"width":7.0, "height":7.0},
                        "rastsize":{"width":750, "height":750},
                        "annotations":annotations } )
    drawcmnds.append( { "action":"exit" } )

    # start PyQt
    app = QApplication(["PipedViewerPQ"])
    # create a PipedViewerPQ in this process
    cmndrecvpipe, cmndsendpipe = Pipe(False)
    rspdrecvpipe, rspdsendpipe = Pipe(False)
    viewer = PipedViewerPQ(cmndrecvpipe, rspdsendpipe)
    # create a command submitter dialog
    tester = _CommandSubmitterPQ(viewer, cmndsendpipe,
                                 rspdrecvpipe, drawcmnds)
    tester.show()
    # let it all run
    result = app.exec_()
    if result != 0:
        sys.exit(result)

if __name__ == "__main__":
    _test_pipedviewerpq()



'''
Regridders designed for use in PyFerret, especially for Python external
functions for PyFerret.  Includes the singleton class ESMPControl to
safely start and stop ESMP once, and only once, in a Python session.

@author: Karl Smith
'''

import numpy

# Import classes given in modules in this package so they are all seen here.
try:
    from esmpcontrol import ESMPControl
    from regrid2d import CurvRectRegridder
    from regrid3d import CurvRect3DRegridder
except ImportError:
    # No ESMP, but do not raise an error until attempting to actually use it
    pass


def quadCornersFrom3D(ptx3d, pty3d):
    '''
    Converts 3D formats of curvilinear corner coordinates to 2D format.

    In 3D formats, quad[i,j] is formed by consecutive corner points
    in the sequence:
        (ptx3d[i,j,0], pty3d[i,j,0]),
        (ptx3d[i,j,1], pty3d[i,j,1]),
        (ptx3d[i,j,2], pty3d[i,j,2]),
        (ptx3d[i,j,3], pty3d[i,j,3]),
        (ptx3d[i,j,0], pty3d[i,j,0]),
    thus graphically represented, using the third indices, as:
          3 --- 2        1 --- 2
         /     /   or   /     /
        0 --- 1        0 --- 3
    Neighboring quadrilaterals must share edges, thus their 3D corner
    coordinates values must also satisfy:
        quad[i+1,j,0] is the same as quad[i,j,1] or quad[i,j,3]
        quad[i,j+j,0] is the same as quad[i,j,3] or quad[i,j,1]
        quad[i+1,j+1,0] is the same as quad[i,j,2]

    In the 2D format, quad[i,j] is former by consecutive corner points
    in the sequence:
        (ptx[i,   j],   pty[i,   j]) ,
        (ptx[i+1, j],   pty[i+1, j]) ,
        (ptx[i+1, j+1], pty[i+1, j+1]) ,
        (ptx[i,   j+1], pty[i,   j+1]) ,
        (ptx[i,   j],   pty[i,   j])
    The shape value of the first two dimensions for the 2D corner point
    arrays is one larger than the corresponding shape value for the 3D
    corner points.

    Arguments:
        ptx3d - 3D quadrilateral corner X coordinates
        pty3d - 3D quadrilateral corner Y coordinates
    Returns:
        (ptx, pty) where
        ptx - 2D quadrilateral corner X coordinates
        pty - 2D quadrilateral corner Y coordiantes
    Raises:
        ValueError if ptx3d or pty3d do not fit one of the expected formats
    '''
    # relative and absolute tolerances to be used in numpy.allclose
    rtol = 1.0E-4
    atol = 1.0E-6

    if ptx3d.shape != pty3d.shape:
        raise ValueError("ptx3d and pty3d in quadCornersFrom3D must have the same shape ")
    corners_shape = (ptx3d.shape[0] + 1, ptx3d.shape[1] + 1)

    if not (numpy.allclose(ptx3d[1:, 1:, 0], ptx3d[:-1, :-1, 2], rtol, atol) and \
            numpy.allclose(pty3d[1:, 1:, 0], pty3d[:-1, :-1, 2], rtol, atol)):
        raise ValueError("Unexpected ptx3d, pty3d values in quadCornersFrom3D")

    # Check if corners are: 
    #     3 --- 2        1 --- 2 
    #    /     /   or   /     / 
    #   0 --- 1        0 --- 3 
    # and assign appropriately. 
    if numpy.allclose(ptx3d[1:, :, 0], ptx3d[:-1, :, 1], rtol, atol) and \
       numpy.allclose(pty3d[1:, :, 0], pty3d[:-1, :, 1], rtol, atol) and \
       numpy.allclose(ptx3d[:, 1:, 0], ptx3d[:, :-1, 3], rtol, atol) and \
       numpy.allclose(pty3d[:, 1:, 0], pty3d[:, :-1, 3], rtol, atol):
        ptx = numpy.empty(corners_shape, dtype=numpy.float64)
        ptx[0, 0] = ptx3d[0, 0, 0]
        ptx[0, 1:] = ptx3d[0, :, 3]
        ptx[1:, 0] = ptx3d[:, 0, 1]
        ptx[1:, 1:] = ptx3d[:, :, 2]
        pty = numpy.empty(corners_shape, dtype=numpy.float64)
        pty[0, 0] = pty3d[0, 0, 0]
        pty[0, 1:] = pty3d[0, :, 3]
        pty[1:, 0] = pty3d[:, 0, 1]
        pty[1:, 1:] = pty3d[:, :, 2]
    elif numpy.allclose(ptx3d[1:, :, 0], ptx3d[:-1, :, 3], rtol, atol) and \
         numpy.allclose(pty3d[1:, :, 0], pty3d[:-1, :, 3], rtol, atol) and \
         numpy.allclose(ptx3d[:, 1:, 0], ptx3d[:, :-1, 1], rtol, atol) and \
         numpy.allclose(pty3d[:, 1:, 0], pty3d[:, :-1, 1], rtol, atol):
        ptx = numpy.empty(corners_shape, dtype=numpy.float64)
        ptx[0, 0] = ptx3d[0, 0, 0]
        ptx[0, 1:] = ptx3d[0, :, 1]
        ptx[1:, 0] = ptx3d[:, 0, 3]
        ptx[1:, 1:] = ptx3d[:, :, 2]
        pty = numpy.empty(corners_shape, dtype=numpy.float64)
        pty[0, 0] = pty3d[0, 0, 0]
        pty[0, 1:] = pty3d[0, :, 1]
        pty[1:, 0] = pty3d[:, 0, 3]
        pty[1:, 1:] = pty3d[:, :, 2]
    else:
        raise ValueError("Unexpected ptx3d, pty3d values in quadCornersFrom3D")

    return (ptx, pty)


def quadCentroids(ptx, pty):
    '''
    Returns the centroids of quadrilaterals using the Surveyor's formula.
    The quadilaterals are defined from the 2D corner point numpy arrays
    ptx and pty, which must have the same shape, where quad[i,j] is formed
    by joined consecutive x,y points in the sequence:
    ( ptx[i,   j],   pty[i,   j] ) ,
    ( ptx[i+1, j],   pty[i+1, j] ) ,
    ( ptx[i+1, j+1], pty[i+1, j+1] ) ,
    ( ptx[i,   j+1], pty[i,   j+1] ) ,
    ( ptx[i,   j],   pty[i,   j] )
    The ctrx and ctry arrays returned will have a shape one smaller in the
    first two dimensions compared to ptx and pty.
    '''
    if ptx.shape != pty.shape:
        raise ValueError("ptx and pty in quadCentroids must have the same shape")

    side0010 = ptx[:-1, :-1] * pty[1:,  :-1] - ptx[1:,  :-1] * pty[:-1, :-1]
    side1011 = ptx[1:,  :-1] * pty[1:,  1:]  - ptx[1:,  1:]  * pty[1:,  :-1]
    side1101 = ptx[1:,  1:]  * pty[:-1, 1:]  - ptx[:-1, 1:]  * pty[1:,  1:]
    side0100 = ptx[:-1, 1:]  * pty[:-1, :-1] - ptx[:-1, :-1] * pty[:-1, 1:]

    area = 0.5 * (side0010 + side1011 + side1101 + side0100)

    ctrx  = (ptx[:-1, :-1] + ptx[1:,  :-1]) * side0010
    ctrx += (ptx[1:,  :-1] + ptx[1:,  1:])  * side1011
    ctrx += (ptx[1:,  1:]  + ptx[:-1, 1:])  * side1101
    ctrx += (ptx[:-1, 1:]  + ptx[:-1, :-1]) * side0100
    ctrx /= 6.0 * area

    ctry  = (pty[:-1, :-1] + pty[1:,  :-1]) * side0010
    ctry += (pty[1:,  :-1] + pty[1:,  1:])  * side1011
    ctry += (pty[1:,  1:]  + pty[:-1, 1:])  * side1101
    ctry += (pty[:-1, 1:]  + pty[:-1, :-1]) * side0100
    ctry /= 6.0 * area

    return (ctrx, ctry)




"""
Helper functions for pyferret shapefile external functions.
"""

from __future__ import print_function

import numpy
import shapefile
import os
import os.path
import pyferret.fershp.mapprj


def createprjfile(shapefile_mapprj, shapefile_name):
    """
    Creates a map projection (.prj) file for a shapefile.

    Arguments:
        shapefile_mapprj - either the common name or the WKT
                           description of the map projection;
                           if None or blank, "WGS 84" is used.
        shapefile_name   - name of the shapefile; any filename
                           extensions are ignored.
    Raises:
        ValueError if the map projection is invalid.
    """
    # If the string given looks like a WKT description, just use it;
    # otherwise, try to convert the name into a description.
    if (not shapefile_mapprj) or shapefile_mapprj.isspace():
        prj_descript = pyferret.fershp.mapprj.name_to_descript("WGS 84")
    elif shapefile_mapprj.startswith('GEOGCS["') or \
         shapefile_mapprj.startswith('PROJCS["'):
        prj_descript = shapefile_mapprj
    else:
        prj_descript = pyferret.fershp.mapprj.name_to_descript(shapefile_mapprj)
    (sfname, ext) = os.path.splitext(shapefile_name)
    prjfile = file("%s.prj" % sfname, "w")
    print(prj_descript, file=prjfile)
    prjfile.close()


def quadxycentroids(xvals, yvals):
    """
    Returns the centroids of X,Y-quadrilaterals whose vertices
    are given by xvals and yvals.

    Arguments:
        xvals - 2D array of X values of the quadrilateral vertices
        yvals - 2D array of Y values of the quadrilateral vertices
        Quadrilaterals are defined by the (xvals, yvals) of
        [i,j] -> [i,j+1] -> [i+1,j+1] -> [i+1,j] -> [i,j]

    Returns:
        Two 2D arrays of X values and Y values of the quadrilateral
        centroids.  The size of each dimension is decreased by one.

    Raises:
        ValueError if the arguments are invalid
    """
    xarray = numpy.asarray(xvals, dtype=float)
    yarray = numpy.asarray(yvals, dtype=float)
    if len(xarray.shape) < 2:
        raise ValueError("xvals and yvals must be (at least) two dimensional")
    if xarray.shape != yarray.shape:
        raise ValueError("xvals and yvals must have the same dimensions")
    sixareas  = xarray[:-1,:-1] * yarray[:-1,1:]  - xarray[:-1,1:]  * yarray[:-1,:-1]
    sixareas += xarray[:-1,1:]  * yarray[1:,1:]   - xarray[1:,1:]   * yarray[:-1,1:]
    sixareas += xarray[1:,1:]   * yarray[1:,:-1]  - xarray[1:,:-1]  * yarray[1:,1:]
    sixareas += xarray[1:,:-1]  * yarray[:-1,:-1] - xarray[:-1,:-1] * yarray[1:,:-1]
    sixareas *= 3.0
    cenxs  = ( xarray[:-1,:-1] * yarray[:-1,1:]  - xarray[:-1,1:]  * yarray[:-1,:-1] ) \
             * ( xarray[:-1,:-1] + xarray[:-1,1:]  )
    cenxs += ( xarray[:-1,1:]  * yarray[1:,1:]   - xarray[1:,1:]   * yarray[:-1,1:] ) \
             * ( xarray[:-1,1:]  + xarray[1:,1:]   )
    cenxs += ( xarray[1:,1:]   * yarray[1:,:-1]  - xarray[1:,:-1]  * yarray[1:,1:] ) \
             * ( xarray[1:,1:]   + xarray[1:,:-1]  )
    cenxs += ( xarray[1:,:-1]  * yarray[:-1,:-1] - xarray[:-1,:-1] * yarray[1:,:-1] ) \
             * ( xarray[1:,:-1]  + xarray[:-1,:-1] )
    cenxs /= sixareas
    cenys  = ( xarray[:-1,:-1] * yarray[:-1,1:]  - xarray[:-1,1:]  * yarray[:-1,:-1] ) \
             * ( yarray[:-1,:-1] + yarray[:-1,1:]  )
    cenys += ( xarray[:-1,1:]  * yarray[1:,1:]   - xarray[1:,1:]   * yarray[:-1,1:] ) \
             * ( yarray[:-1,1:]  + yarray[1:,1:]   )
    cenys += ( xarray[1:,1:]   * yarray[1:,:-1]  - xarray[1:,:-1]  * yarray[1:,1:] ) \
             * ( yarray[1:,1:]   + yarray[1:,:-1]  )
    cenys += ( xarray[1:,:-1]  * yarray[:-1,:-1] - xarray[:-1,:-1] * yarray[1:,:-1] ) \
             * ( yarray[1:,:-1]  + yarray[:-1,:-1] )
    cenys /= sixareas
    return (cenxs, cenys)


def quadxycenters(xvals, yvals):
    """
    Returns the average centers of X,Y-quadrilaterals whose vertices
    are given by xvals and yvals.

    Arguments:
        xvals - 2D array of X values of the quadrilateral vertices
        yvals - 2D array of Y values of the quadrilateral vertices
        Quadrilaterals are defined by the (xvals, yvals) of
        [i,j] -> [i,j+1] -> [i+1,j+1] -> [i+1,j] -> [i,j]

    Returns:
        Two 2D arrays of X values and Y values of the quadrilateral
        average centers.  The size of each dimension is decreased by one.

    Raises:
        ValueError if the arguments are invalid
    """
    xarray = numpy.asarray(xvals, dtype=float)
    yarray = numpy.asarray(yvals, dtype=float)
    if len(xarray.shape) < 2:
        raise ValueError("xvals and yvals must be (at least) two dimensional")
    if xarray.shape != yarray.shape:
        raise ValueError("xvals and yvals must have the same dimensions")
    cenxs  = 0.25 * ( xarray[:-1,:-1] + xarray[:-1,1:] + xarray[1:,1:] + xarray[1:,:-1] )
    cenys  = 0.25 * ( yarray[:-1,:-1] + yarray[:-1,1:] + yarray[1:,1:] + yarray[1:,:-1] )
    return (cenxs, cenys)


def addquadxyvalues(sfwriter, pt0, pt1, pt2, pt3, zcoord, vals):
    """
    Adds a quadrilateral shape to sfwriter defined by the X,Y vertices
    pt0 - pt1 - pt2 - pt3 - pt0, and possibly the common Z coordinate
    (or array of Z coordinates) zcoord, along with the associated values
    in vals.

    Arguments:
       sfwriter - the shapefile.Writer object to add the shape and values to
       pt1, pt2,
       pt3, pt4 - the (X,Y) numeric coordinates of the vertices of the simple
                  quadrilateral; in sequence, but not necessarily the correct
                  winding.  Any coordinates after the first two in each point
                  are ignored.
       zcoord   - the numeric Z coordinate or array of numeric Z coordinates
                  for this quadrilateral; may be None
       vals     - the list of values to be associated with this shape.  The
                  fields for these values must already have been created in
                  sfwriter.

    Note: the winding of the quadrilateral is determined only using the X and
          Y coordinates, even when multiple Z coordinates are provided.
    """
    # Get the correct polygon type
    if zcoord != None:
        shapetype = shapefile.POLYGONZ
    else:
        shapetype = shapefile.POLYGON
    x0 = float(pt0[0]); y0 = float(pt0[1])
    x1 = float(pt1[0]); y1 = float(pt1[1])
    x2 = float(pt2[0]); y2 = float(pt2[1])
    x3 = float(pt3[0]); y3 = float(pt3[1])
    # Compute 2 * signed area of this simple quadrilateral
    dqarea  = x0 * y1 - x1 * y0
    dqarea += x1 * y2 - x2 * y1
    dqarea += x2 * y3 - x3 * y2
    dqarea += x3 * y0 - x0 * y3
    # Create the correctly ordered array of coordinates for this single shape part
    part = [ ]
    if dqarea < 0.0:
        # negative means clockwise which is desired direction
        part.append([ x0, y0 ])
        part.append([ x1, y1 ])
        part.append([ x2, y2 ])
        part.append([ x3, y3 ])
        part.append([ x0, y0 ])
    else:
        # positive means counterclockwise so reverse ordering
        part.append([ x0, y0 ])
        part.append([ x3, y3 ])
        part.append([ x2, y2 ])
        part.append([ x1, y1 ])
        part.append([ x0, y0 ])
    # Append the Z coordinate(s) if given
    if zcoord != None:
        try:
            # First try it as a single Z coordinate for all points
            z = float(zcoord)
            for pt in part:
                pt.append(z)
        except TypeError:
            if len(zcoord) != 4:
                raise ValueError("zcoord must be None, a single value, or a list of four values")
            # Assume it is an array of numbers
            if dqarea < 0.0:
                zvals = list(zcoord) + [ zcoord[0] ]
            else:
                zvals = [ zcoord[0] ] + list(zcoord[::-1])
            for (pt, z) in zip(part, zvals):
                pt.append(float(z))
    # Add the shape
    sfwriter.poly([ part, ], shapetype)
    # Add the values for this shape
    sfwriter.record(*vals)


#
# The following is only for testing this module from the command line
#
if __name__ == "__main__":

    shapefilename = "testsf"
    wgs84_descript = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]]'
    wgs84upsnorth_descript = 'PROJCS["WGS 84 / UPS North",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Polar_Stereographic"],PARAMETER["latitude_of_origin",90],PARAMETER["central_meridian",0],PARAMETER["scale_factor",0.994],PARAMETER["false_easting",2000000],PARAMETER["false_northing",2000000],UNIT["metre",1]]'

    # Test createprjfile
    createprjfile(None, "%s.jnk" % shapefilename)
    prjfile = file("%s.prj" % shapefilename, "r")
    datalines = prjfile.readlines()
    prjfile.close()
    if len(datalines) != 1:
        raise ValueError("None for mapprj: more than one line given in the .prj file")
    descript = datalines[0].strip()
    if descript != wgs84_descript:
        raise ValueError("None for mapprj:\n" \
                         "    expect: %s\n" \
                         "    found:  %s" % (wgs84_descript, descript))
    del prjfile, datalines, descript

    prjjunk = 'GEOGCS["Junk",DATUM["Junk"]]'
    createprjfile(prjjunk, "%s.jnk" % shapefilename)
    prjfile = file("%s.prj" % shapefilename, "r")
    datalines = prjfile.readlines()
    prjfile.close()
    if len(datalines) != 1:
        raise ValueError("Junk for mapprj: more than one line given in the .prj file")
    descript = datalines[0].strip()
    if descript != prjjunk:
        raise ValueError("Junk for mapprj:\n" \
                         "    expect: %s\n" \
                         "    found:  %s" % (prjjunk, descript))
    del prjjunk, prjfile, datalines, descript

    createprjfile("WGS 84 / UPS North", "%s.jnk" % shapefilename)
    prjfile = file("%s.prj" % shapefilename, "r")
    datalines = prjfile.readlines()
    prjfile.close()
    if len(datalines) != 1:
        raise ValueError("'WGS 84 / UPS North' for mapprj: more than one line given in the .prj file")
    descript = datalines[0].strip()
    if descript != wgs84upsnorth_descript:
        raise ValueError("'WGS 84 / UPS North' for mapprj:\n" \
                         "    expect: %s\n" \
                         "    found:  %s" % (wgs84upsnorth_descript, descript))
    del prjfile, datalines, descript
    print("createprjfile: SUCCESS")

    # Test quadxycentroids
    xvals = ( ( 0, 1 ), ( 3, 4 ) )
    yvals = ( ( 0, 2 ), ( 1, 3 ) )
    expectx = [ [ 2.0 ] ]
    expecty = [ [ 1.5 ] ]
    (centx, centy) = quadxycentroids(xvals, yvals)
    if not numpy.allclose(centx, expectx):
        raise ValueError("Centroid X values: expected %s; found %s" % \
                          (str(expectx), str(centx)))
    if not numpy.allclose(centy, expecty):
        raise ValueError("Centroid Y values: expected %s; found %s" % \
                          (str(expecty), str(centy)))
    del xvals, yvals, expectx, expecty, centx, centy
    xvals = ( ( 0, 1 ), ( 2, 3 ) )
    yvals = ( ( 0, 2 ), ( 1, 5 ) )
    expectx = [ [ 39.0 / 24.0 ] ]
    expecty = [ [ 49.0 / 24.0 ] ]
    (centx, centy) = quadxycentroids(xvals, yvals)
    if not numpy.allclose(centx, expectx):
        raise ValueError("Centroid X values: expected %s; found %s" % \
                          (str(expectx), str(centx)))
    if not numpy.allclose(centy, expecty):
        raise ValueError("Centroid Y values: expected %s; found %s" % \
                          (str(expecty), str(centy)))
    del xvals, yvals, expectx, expecty, centx, centy
    print("quadxycentroids: SUCCESS")

    # Test quadxycenters
    xvals = ( ( 0, 1 ), ( 3, 4 ) )
    yvals = ( ( 0, 2 ), ( 1, 3 ) )
    expectx = [ [ 2.0 ] ]
    expecty = [ [ 1.5 ] ]
    (centx, centy) = quadxycenters(xvals, yvals)
    if not numpy.allclose(centx, expectx):
        raise ValueError("Centroid X values: expected %s; found %s" % \
                          (str(expectx), str(centx)))
    if not numpy.allclose(centy, expecty):
        raise ValueError("Centroid Y values: expected %s; found %s" % \
                          (str(expecty), str(centy)))
    del xvals, yvals, expectx, expecty, centx, centy
    xvals = ( ( 0, 1 ), ( 2, 3 ) )
    yvals = ( ( 0, 2 ), ( 1, 5 ) )
    expectx = [ [ 1.5 ] ]
    expecty = [ [ 2.0 ] ]
    (centx, centy) = quadxycenters(xvals, yvals)
    if not numpy.allclose(centx, expectx):
        raise ValueError("Centroid X values: expected %s; found %s" % \
                          (str(expectx), str(centx)))
    if not numpy.allclose(centy, expecty):
        raise ValueError("Centroid Y values: expected %s; found %s" % \
                          (str(expecty), str(centy)))
    del xvals, yvals, expectx, expecty, centx, centy
    print("quadxycenters: SUCCESS")

    # Test addquadxyvalues
    coords = [ [0.0, 0.0], [1.0, 0.0], [1.0, -1.0], [2.0, 1.0] ]
    zval = [ -5.34, -4.23, -3.12, -2.01 ]
    vals = [ 3.28573, 7.46952 ]
    expectedxy = [ coords[0], coords[3], coords[2], coords[1], coords[0] ]
    expectedz = [ zval[0], zval[3], zval[2], zval[1], zval[0] ]
    # Create the shapefile
    sfwriter = shapefile.Writer(shapefile.POLYGONZ)
    sfwriter.field("VAL0", "N", 20, 7)
    sfwriter.field("VAL1", "N", 20, 7)
    # Add the shape and values and save the shapefile
    addquadxyvalues(sfwriter, coords[0], coords[1], coords[2], coords[3], zval, vals)
    sfwriter.save(shapefilename)
    del zval, coords, sfwriter
    # Read the shapefile and check the shape and values
    sfreader = shapefile.Reader(shapefilename)
    shapes = sfreader.shapes()
    if len(shapes) != 1:
        raise ValueError("Expected one shape; found %d" % len(shapes))
    if shapes[0].shapeType != shapefile.POLYGONZ:
        raise ValueError("Expected shapetype %d; found %d" % \
                         (shapefile.POLYGONZ, shapes[0].shapeType))
    if not numpy.allclose(shapes[0].points, expectedxy):
        raise ValueError("Expected (X,Y) coordinates %s; found %s" % \
                         (str(expectedxy), str(shapes[0].points)))
    if not numpy.allclose(shapes[0].z, expectedz):
        raise ValueError("Expected Z coordinates %s; found %s" % \
                         (str(expectedz), str(shapes[0].z)))
    records = sfreader.records()
    if len(records) != 1:
        raise ValueError("Expected one set of records; found %d" % len(records))
    if not numpy.allclose(records[0], vals):
        raise ValueError("Expected record values %s; found %s" % \
                         (str(vals), str(records[0])))
    del expectedxy, expectedz, sfreader, shapes, records
    os.remove("%s.dbf" % shapefilename)
    os.remove("%s.shp" % shapefilename)
    os.remove("%s.shx" % shapefilename)
    os.remove("%s.prj" % shapefilename)

    print("addquadxyvalues: SUCCESS")




"""
Performs a two-sided Kolmogorov-Smirnov test that two samples
come from the same continuous probability distribution.
"""

from __future__ import print_function

import numpy
import pyferret
import scipy.stats


def ferret_init(id):
    """
    Initialization for the stats_kstest2 PyEF
    """
    axes_values = [ pyferret.AXIS_DOES_NOT_EXIST ] * pyferret.MAX_FERRET_NDIM
    axes_values[0] = pyferret.AXIS_CUSTOM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 2,
                "descript": "Returns two-sided Kolmogorov-Smirnov test stat. and prob. " \
                            "that two samples comes from the same prob. distrib.",
                "axes": axes_values,
                "argnames": ( "SAMPLEA", "SAMPLEB", ),
                "argdescripts": ( "First sample data array",
                                  "Second sample data array", ),
                "argtypes": ( pyferret.FLOAT_ARRAY, pyferret.FLOAT_ARRAY, ),
                "influences": ( false_influences, false_influences, ),
              }
    return retdict


def ferret_custom_axes(id):
    """
    Define custom axis of the stats_kstest2 Ferret PyEF
    """
    axis_defs = [ None ] * pyferret.MAX_FERRET_NDIM
    axis_defs[0] = ( 1, 2, 1, "KS,P", False )
    return axis_defs


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Performs a two-sided Kolmogorov-Smirnov test that two samples come
    from the same continuous probability distribution.  The samples are
    given in inputs[0] and inputs[1].  The test statistic value and
    two-tailed probability are returned in result.  Undefined data given
    in each sample are removed (independently from each other) before
    performing the test.  Note that the samples do not need to be the
    same size; thus there are no restrictions on the relative dimensions
    of sample arrays.
    """
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    sampa = inputs[0][goodmask]
    badmask = ( numpy.fabs(inputs[1] - inpbdfs[1]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[1]))
    goodmask = numpy.logical_not(badmask)
    sampb = inputs[1][goodmask]
    fitparams = scipy.stats.ks_2samp(sampa, sampb)
    result[:] = resbdf
    # Kolmogorov-Smirnov test statistic
    result[0] = fitparams[0]
    # probability
    result[1] = fitparams[1]


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init and ferret_custom_axes do not have problems
    info = ferret_init(0)
    info = ferret_custom_axes(0)

    # Get two random samples from the same distribution
    ydim = 250
    zdim = 150
    distf = scipy.stats.weibull_min(2.0, 5.0)
    sampa = distf.rvs(ydim * zdim)
    sampb = distf.rvs(ydim * zdim)

    # Get a distribution from a different distribution
    sampu = scipy.stats.uniform(loc=7.0, scale=5.0).rvs(ydim * zdim)

    # setup for the call to ferret_compute
    inpbdfs = numpy.array([-9999.0, -8888.0], dtype=numpy.float64)
    resbdf = numpy.array([-7777.0], dtype=numpy.float64)
    arraya = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    arrayb = numpy.empty((ydim, 1, 1, zdim, 1, 1), dtype=numpy.float64, order='F')
    arrayu = numpy.empty((ydim, 1, 1, zdim, 1, 1), dtype=numpy.float64, order='F')
    index = 0
    for j in range(ydim):
        for k in range(zdim):
            if (index % 23) == 3:
                arraya[0, j, k, 0, 0, 0] = inpbdfs[0]
            else:
                arraya[0, j, k, 0, 0, 0] = sampa[index]
            if (index % 53) == 3:
                arrayb[j, 0, 0, k, 0, 0] = inpbdfs[1]
                arrayu[j, 0, 0, k, 0, 0] = inpbdfs[1]
            else:
                arrayb[j, 0, 0, k, 0, 0] = sampb[index]
                arrayu[j, 0, 0, k, 0, 0] = sampu[index]
            index += 1
    resultb = -6666.0 * numpy.ones((2, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    resultu = -6666.0 * numpy.ones((2, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute with the samples from the same distribution and check the results
    ferret_compute(0, resultb, resbdf, (arraya, arrayb), inpbdfs)
    resultb = resultb.reshape(-1)
    print("from same dist result: %s" % str(resultb))
    if (resultb[0] < 0.00) or (resultb[0] > 0.01) or \
       (resultb[1] < 0.10) or (resultb[1] > 1.00):
        raise ValueError("Unexpected result")

    # call ferret_compute with data from different distributions and check the results
    ferret_compute(0, resultu, resbdf, (sampa, sampu), inpbdfs)
    resultu = resultu.reshape(-1)
    print("from diff dist result: %s" % str(resultu))
    if (resultu[0] < 0.98) or (resultu[0] > 1.00) or \
       (resultu[1] < 0.00) or (resultu[1] > 0.01):
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import numpy

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> nogrid = pyferret.FerGrid()")
nogrid = pyferret.FerGrid()
print(">>> print repr(nogrid)")
print(repr(nogrid))
print(">>> dir(nogrid)")
dir(nogrid)

print(">>> coads = pyferret.FerDSet('coads_climatology')")
coads = pyferret.FerDSet('coads_climatology')
print(">>> coads.sst.load()")
coads.sst.load()
print(">>> coadsgrid = coads.sst.grid")
coadsgrid = coads.sst.grid
print(">>> print repr(coadsgrid)")
print(repr(coadsgrid))
print(">>> nogrid == coadsgrid")
nogrid == coadsgrid
print(">>> nogrid != coadsgrid")
nogrid != coadsgrid

print(">>> dupgrid = coadsgrid.copy()")
dupgrid = coadsgrid.copy()
print(">>> dupgrid is coadsgrid")
dupgrid is coadsgrid
print(">>> dupgrid == coadsgrid")
dupgrid == coadsgrid
print(">>> dupgrid.axes[0] is coadsgrid.axes[0]")
dupgrid.axes[0] is coadsgrid.axes[0]
print(">>> dupgrid.axes[0] == coadsgrid.axes[0]")
dupgrid.axes[0] == coadsgrid.axes[0]
print(">>> dupgrid.axes[1] is coadsgrid.axes[1]")
dupgrid.axes[1] is coadsgrid.axes[1]
print(">>> dupgrid.axes[1] == coadsgrid.axes[1]")
dupgrid.axes[1] == coadsgrid.axes[1]
print(">>> dupgrid.axes[2] is coadsgrid.axes[2]")
dupgrid.axes[2] is coadsgrid.axes[2]
print(">>> dupgrid.axes[2] == coadsgrid.axes[2]")
dupgrid.axes[2] == coadsgrid.axes[2]
print(">>> dupgrid.axes[3] is coadsgrid.axes[3]")
dupgrid.axes[0] is coadsgrid.axes[0]
print(">>> dupgrid.axes[3] == coadsgrid.axes[3]")
dupgrid.axes[3] == coadsgrid.axes[3]
print(">>> dupgrid.axes[4] is coadsgrid.axes[4]")
dupgrid.axes[4] is coadsgrid.axes[4]
print(">>> dupgrid.axes[4] == coadsgrid.axes[4]")
dupgrid.axes[4] == coadsgrid.axes[4]
print(">>> dupgrid.axes[5] is coadsgrid.axes[5]")
dupgrid.axes[0] is coadsgrid.axes[0]
print(">>> dupgrid.axes[5] == coadsgrid.axes[5]")
dupgrid.axes[5] == coadsgrid.axes[5]

print(">>> freqax = pyferret.FerAxis(axtype=pyferret.AXISTYPE_CUSTOM,coords=numpy.arange(1,13,0.5),unit='freqnum',name='frequencies')")
freqax = pyferret.FerAxis(axtype=pyferret.AXISTYPE_CUSTOM,coords=numpy.arange(1,13,0.5),unit='freqnum',name='frequencies')
print(">>> print repr(freqax)")
print(repr(freqax))
print(">>> freqgrid = coadsgrid.copy(ax=pyferret.T_AXIS,newax=None).copy(name='freqgrid',ax=pyferret.E_AXIS,newax=freqax)")
freqgrid = coadsgrid.copy(ax=pyferret.T_AXIS,newax=None).copy(name='freqgrid',ax=pyferret.E_AXIS,newax=freqax)
print(">>> print repr(freqgrid)")
print(repr(freqgrid))
print(">>> freqgrid.axes[0] == coadsgrid.axes[0]")
freqgrid.axes[0] == coadsgrid.axes[0]
print(">>> freqgrid.axes[1] == coadsgrid.axes[1]")
freqgrid.axes[1] == coadsgrid.axes[1]
print(">>> freqgrid.axes[2] == coadsgrid.axes[2]")
freqgrid.axes[2] == coadsgrid.axes[2]
print(">>> freqgrid.axes[3] == coadsgrid.axes[3]")
freqgrid.axes[3] == coadsgrid.axes[3]
print(">>> freqgrid.axes[4] == coadsgrid.axes[4]")
freqgrid.axes[4] == coadsgrid.axes[4]
print(">>> freqgrid.axes[5] == coadsgrid.axes[5]")
freqgrid.axes[5] == coadsgrid.axes[5]




'''
Represents Ferret axes in Python.
'''

import numbers
import time
import numpy
import pyferret

# set of valid axis types
_VALID_AXIS_TYPES = frozenset( (pyferret.AXISTYPE_LONGITUDE, 
                                pyferret.AXISTYPE_LATITUDE, 
                                pyferret.AXISTYPE_LEVEL, 
                                pyferret.AXISTYPE_TIME, 
                                pyferret.AXISTYPE_CUSTOM, 
                                pyferret.AXISTYPE_ABSTRACT, 
                                pyferret.AXISTYPE_NORMAL) )

# Supported formats for time.strptime
_TIME_PARSE_FORMATS = ( 
    '%d-%b-%Y %H:%M:%S',
    '%d-%b-%Y %H:%M',
    '%d-%b-%Y',
    '%Y-%m-%dT%H:%M:%S',
    '%Y-%m-%dT%H:%M',
    '%Y-%m-%d %H:%M:%S',
    '%Y-%m-%d %H:%M',
    '%Y-%m-%d', 
)
_TIME_NOYEAR_PARSE_FORMATS = (
    '%d-%b %H:%M:%S',
    '%d-%b',
)


class FerAxis(object):
    '''
    Ferret axis object
    '''

    def __init__(self, coords=None, axtype=None, unit=None, name=None):
        '''
        Describe a Ferret axis using the given information about the axis.
            axtype (int): type of the axis; valid values are
                    pyferret.AXISTYPE_LONGITUDE
                    pyferret.AXISTYPE_LATITUDE
                    pyferret.AXISTYPE_LEVEL
                    pyferret.AXISTYPE_TIME
                    pyferret.AXISTYPE_CUSTOM   (axis unit not recognized by Ferret)
                    pyferret.AXISTYPE_ABSTRACT (axis is unit-less integer values)
                    pyferret.AXISTYPE_NORMAL   (axis is normal to the data)
                if not given but coords is given, AXISTYPE_CUSTOM is used,
                if not given and coords is not given, AXISTYPE_NORMAL is used.
            coords (sequence of numeric): coordinate values of the axis; 
                for an axis that is neither a time axis, an abstract axis, nor normal 
                to the data, a 1-D array of numeric values; 
                for a time axis, an (n,6) 2D array of integers where each time step is 
                formed from the six integers for the day, month, year, hour, minute, 
                and second in the index given by
                    pyferret.TIMEARRAY_DAYINDEX
                    pyferret.TIMEARRAY_MONTHINDEX
                    pyferret.TIMEARRAY_YEARINDEX
                    pyferret.TIMEARRAY_HOURINDEX
                    pyferret.TIMEARRAY_MINUTEINDEX
                    pyferret.TIMEARRAY_SECONDINDEX
                (Thus, coords[t, TIMEARRAY_YEARINDEX] gives the year of time t.)
                Note: a relative time axis will be of type AXISTYPE_CUSTOM, with a unit
                      indicating the starting point, such as 'days since 01-JAN-2000'
                For an abstract axis or an axis normal to the data, this argument is ignored.
            unit (string): unit of the axis; for a time axis, this gives the calendar 
                    as one of
                    pyferret.CALTYPE_360DAY
                    pyferret.CALTYPE_NOLEAP
                    pyferret.CALTYPE_GREGORIAN
                    pyferret.CALTYPE_JULIAN
                    pyferret.CALTYPE_ALLLEAP
                    pyferret.CALTYPE_NONE    (calendar not specified)
                For abstract axes, or axes normal to the data, this argument is ignored.
            name (string): Ferret name for the axis
                For an axis normal to the data, this argument is ignored.
        '''
        # axis type
        if axtype:
            if not axtype in _VALID_AXIS_TYPES:
                raise ValueError('axis type %s is not valid' % str(axtype))
            self._axtype = axtype
        elif coords is not None:
            self._axtype = pyferret.AXISTYPE_CUSTOM
        else:
            self._axtype = pyferret.AXISTYPE_NORMAL
        # axis name
        if name and (self._axtype != pyferret.AXISTYPE_NORMAL):
            if not isinstance(name, str): 
                raise ValueError('axis name %s is not valid' % str(name))
            self._name = name.strip()
        else:
            self._name = ''
        # axis unit
        if unit and (self._axtype != pyferret.AXISTYPE_NORMAL) \
                    and (self._axtype != pyferret.AXISTYPE_ABSTRACT):
            if not isinstance(unit, str): 
                raise ValueError('axis unit %s is not valid' % str(unit))
            self._unit = unit.strip()
        else:
            self._unit = ''
        # axis coordinates
        if (coords is not None) and (self._axtype != pyferret.AXISTYPE_NORMAL) \
                                and (self._axtype != pyferret.AXISTYPE_ABSTRACT):
            if self._axtype == pyferret.AXISTYPE_TIME:
                try:
                    self._coords = numpy.array(coords, dtype=numpy.int32, copy=True)
                except ValueError:
                    raise ValueError('coordinates for a time axis is not an integer array')
                if self._coords.ndim != 2:
                    raise ValueError('coordinates for a time axis is not a 2-D array')
                if self._coords.shape[1] != 6:
                    raise ValueError('second dimenstion of coordinates for a time axis is not 6')
            else:
                try:
                    self._coords = numpy.array(coords, dtype=numpy.float64, copy=True)
                except ValueError:
                    raise ValueError('coordinates for an axis is not a numeric array')
                if self._coords.ndim != 1:
                    raise ValueError('coordinates for a lon/lat/level/custom axis is not a 1-D array' % k)
        else:
            self._coords = None


    def __repr__(self):
        '''
        Representation to recreate this FerAxis
        '''
        # Not elegant, but will do
        infostr = "FerAxis(coords=" + repr(self._coords) + \
                  ", axtype=" + repr(self._axtype) + \
                  ", unit='" + self._unit + \
                  "', name='" + self._name + "')"
        return infostr


    def __eq__(self, other):
        '''
        Two FerAxis objects are equal is all their contents are the same.  
        All string values are compared case-insensitive.
        '''
        if not isinstance(other, FerAxis):
            return NotImplemented
        # _axtype is an integer
        if self._axtype != other._axtype:
            return False
        # _name is a string
        if self._name.upper() != other._name.upper():
            return False
        # _unit is a string
        if self._unit.upper() != other._unit.upper():
            return False
        # _coords is an ndarray or None
        if (self._coords is None) and (other._coords is None):
            return True
        if (self._coords is None) or (other._coords is None):
            return False
        if not numpy.allclose(self._coords, other._coords):
            return False
        return True


    def __ne__(self, other):
        '''
        Two FerAxis obect are not equal is any of their contents are not 
        the same.  All string values are compared case-insensitive.
        '''
        if not isinstance(other, FerAxis):
            return NotImplemented
        return not self.__eq__(other)


    def __getitem__(self, name):
        '''
        Return the axis type (if name='axtype'), unit (if name='unit'), 
        name (if name='name'), or a copy of the coordinates (if name='coords')
        '''
        if name == 'axtype':
            return self.getaxtype()
        if name == 'unit':
            return self.getunit()
        if name == 'name':
            return self.getname()
        if name == 'coords':
            return self.getcoords()
        raise KeyError("unknown key '%s'" % str(name))


    def __getattr__(self, name):
        '''
        Return the axis type (if name='axtype'), unit (if name='unit'), 
        name (if name='name'), or a copy of the coordinates (if name='coords')
        Note that this method is only called when the parent object 
        does not have an attribute with this name.
        '''
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError("unknown attribute '%s'" % name)


    def __dir__(self):
        '''
        Returns a list of known attributes, including those added 
        by the __getattr__ method.
        '''
        mydir = [ 'axtype', 'coords', 'name', 'unit' ]
        mydir.extend( dir(super(FerAxis, self)) )
        return mydir


    def copy(self):
        '''
        Returns a copy of this FerAxis object.  The FerAxis object returned
        does not share any mutable values (namely, the coordinates array)
        with this FerAxis object.
        '''
        # __init__ forces a copy of the coordinates array
        duplicate = FerAxis(axtype=self._axtype, coords=self._coords,
                            unit=self._unit, name=self._name)
        return duplicate


    def getaxtype(self):
        '''
        Returns the type of this axis as one of the integer constants
            pyferret.AXISTYPE_LONGITUDE
            pyferret.AXISTYPE_LATITUDE
            pyferret.AXISTYPE_LEVEL
            pyferret.AXISTYPE_TIME
            pyferret.AXISTYPE_CUSTOM   (axis unit not recognized by Ferret)
            pyferret.AXISTYPE_ABSTRACT (axis is unit-less integer values)
            pyferret.AXISTYPE_NORMAL   (axis is normal to the data)
        '''
        return self._axtype


    def getcoords(self):
        '''
        Returns a copy of the coordinates ndarray for this axis, 
        or None if there is no coordinates array for this axis.
        '''
        if self._coords is not None:
            coords = self._coords.copy('A')
        else:
            coords = None
        return coords


    def getunit(self):
        '''
        Returns the unit string for this axis.  May be an empty string.
        '''
        return self._unit


    def getname(self):
        '''
        Returns the name string for this axis.  May be an empty string.
        '''
        return self._name


    @staticmethod
    def _parsegeoslice(geoslice):
        '''
        Parses the contents of the slice attributes, interpreting any geo- or time-references
        and returns a tuple with the resulting interpreted axis type, start, stop, and step values.

           geoslice (slice): slice that can contain georeferences or time references

           returns (axtype, start, stop, step) where:
              axtype is one of:
                  pyferret.AXISTYPE_LONGITUDE  (longitude units detected)
                  pyferret.AXISTYPE_LATITUDE   (latitude units detected)
                  pyferret.AXISTYPE_LEVEL      (level units detected)
                  pyferret.AXISTYPE_TIME       (time units detected)
                  pyferret.AXISTYPE_ABSTRACT   (no units)
              start, stop, and step are:
                  None if the correspond geoslice attribute is not given; otherwise,
                  a list of six numbers if axtype is pyferret.AXISTYPE_TIME, or
                  a number if axtype is not pyferret.AXISTYPE_TIME

        The list of six numbers for time values are ordered according to the indices:
            pyferret.TIMEARRAY_DAYINDEX
            pyferret.TIMEARRAY_MONTHINDEX
            pyferret.TIMEARRAY_YEARINDEX
            pyferret.TIMEARRAY_HOURINDEX
            pyferret.TIMEARRAY_MINUTEINDEX
            pyferret.TIMEARRAY_SECONDINDEX
        For non-time values, the start, stop, and step values are int objects 
            if only if corresponding slice objects were int objects.  Thus, int 
            objects should be interpreted as axis indices and float objects 
            should be interpreted as axis values.
        Raises a ValueError if start and stop indicate different axes; i.e., 
            "10E":"20N" or 10:"20N" or 10:"20-JAN-2000", or if the value contain 
            unrecognized units.  If not a time slice, it is acceptable for step to 
            have no units even when start and stop do.  If a time slice, the step 
            must have a unit of y, d, h, m, or s, which corresponds to year, day, 
            hour, minute, or second; there is no month time step unit.
        Raises a TypeError if geoslice is not a slice or None, or if the values 
            in the slice are not None and cannot be interpreted.
        '''
        if geoslice is None:
            return (pyferret.AXISTYPE_ABSTRACT, None, None, None)
        if not isinstance(geoslice, slice):
            raise TypeError('not a slice object: %s' % repr(geoslice))
        (starttype, start) = FerAxis._parsegeoval(geoslice.start)
        (stoptype, stop) = FerAxis._parsegeoval(geoslice.stop)
        # start and stop types must match (so 10:"25E" also fails)
        if starttype != stoptype:
            raise ValueError('mismatch of units: %s and %s' % (geoslice.start, geoslice.stop))
        axtype = starttype
        if axtype == pyferret.AXISTYPE_TIME:
            (steptype, step) = FerAxis._parsegeoval(geoslice.step, istimestep=True)
            if (step is not None) and (steptype != pyferret.AXISTYPE_TIME):
               raise ValueError('a time unit y, d, h, m, or s must be given with time slice steps')
        else:
            (steptype, step) = FerAxis._parsegeoval(geoslice.step)
            if (steptype != pyferret.AXISTYPE_ABSTRACT) and (steptype != axtype):
               raise ValueError('mismatch of units: %s, %s' % (geoslice.start, geoslice.step))
        return (axtype, start, stop, step)


    @staticmethod
    def _parsegeoval(val, istimestep=False):
        '''
        Parses the value as either a longitude, latitude, level, time, or abstract number.
        If val is a numeric value, the tuple (pyferret.AXISTYPE_ABSTRACT, val) is returned.
        If val is None, the tuple (pyferret.AXISTYPE_ABSTRACT, None) is returned.
        If val is a longitude string (unit E or W when istimestep is false), 
            (pyferret.AXISTYPE_LONGITUDE, fval) is returned where fval 
            is the floating point longitude value.
        If val is a latitude string (unit N or S when istimestep is false), 
            (pyferret.AXISTYPE_LATITUDE, fval) is returned where fval 
            is the floating point latitude value.
        If val is a level string (unit m when istimestep is False), 
            (pyferret.AXISTYPE_LEVEL, fval) is returned where fval 
            is the floating point level value.
        If val is a date and, optionally, time string matching one of the formats given
            in _TIME_PARSE_FORMATS or _TIME_NOYEAR_PARSE_FORMATS, 
            (pyferret.AXISTYPE_TIME, tval) is returned where
            tval is a list of six numbers ordered by the indices:
                pyferret.TIMEARRAY_DAYINDEX
                pyferret.TIMEARRAY_MONTHINDEX
                pyferret.TIMEARRAY_YEARINDEX
                pyferret.TIMEARRAY_HOURINDEX
                pyferret.TIMEARRAY_MINUTEINDEX
                pyferret.TIMEARRAY_SECONDINDEX
        If istimestep is true and val is a time step string (unit y, d, h, m, or s),
            (pyferret.AXISTYPE_TIME, tval) is returned where tval is a list of six values 
            ordered by the above TIMEARRAY indices.  
            Note that m is minutes; there is no month timestep.
        If val is a string of a unitless number, (pyferret.AXISTYPE_ABSTACT, fval) is 
            returned where fval is the floating point value specified by val.
        If val is not numeric or a string, a TypeError is raised.
        If val is a string that cannot be parsed, a ValueError is raised.
        '''
        # if just a number, return it with abstract axis type
        if isinstance(val, numbers.Real):
            return (pyferret.AXISTYPE_ABSTRACT, val)
        # if None or empty, return None with abstract axis type
        if not val:
            return (pyferret.AXISTYPE_ABSTRACT, None)
        if not isinstance(val, str):
            raise TypeError('not a string: %s' % repr(val))
        if not istimestep:
            # not a time *step* - first try parsing as a date/time string using the accepted formats
            for fmt in _TIME_PARSE_FORMATS:
                try:
                    tval = time.strptime(val, fmt)
                    tlist = [ 0, 0, 0, 0, 0, 0 ]
                    tlist[pyferret.TIMEARRAY_DAYINDEX] = tval.tm_mday
                    tlist[pyferret.TIMEARRAY_MONTHINDEX] = tval.tm_mon
                    tlist[pyferret.TIMEARRAY_YEARINDEX] = tval.tm_year
                    tlist[pyferret.TIMEARRAY_HOURINDEX] = tval.tm_hour
                    tlist[pyferret.TIMEARRAY_MINUTEINDEX] = tval.tm_min
                    tlist[pyferret.TIMEARRAY_SECONDINDEX] = tval.tm_sec
                    return (pyferret.AXISTYPE_TIME, tlist)
                except ValueError:
                    pass
            for fmt in _TIME_NOYEAR_PARSE_FORMATS:
                try:
                    tval = time.strptime(val, fmt)
                    tlist = [ 0, 0, 0, 0, 0, 0 ]
                    tlist[pyferret.TIMEARRAY_DAYINDEX] = tval.tm_mday
                    tlist[pyferret.TIMEARRAY_MONTHINDEX] = tval.tm_mon
                    # leave the year as zero - time.strptime assigns 1900
                    tlist[pyferret.TIMEARRAY_HOURINDEX] = tval.tm_hour
                    tlist[pyferret.TIMEARRAY_MINUTEINDEX] = tval.tm_min
                    tlist[pyferret.TIMEARRAY_SECONDINDEX] = tval.tm_sec
                    return (pyferret.AXISTYPE_TIME, tlist)
                except ValueError:
                    pass
        # not a date/time, so parse as a number with possibly a final letter for the unit
        try:
            lastchar = val[-1].upper()
            if (not istimestep) and (lastchar == 'E'): # degrees E
                # make sure the rest is just numeric
                fval = float(val[:-1])
                return(pyferret.AXISTYPE_LONGITUDE, val.upper())
            elif (not istimestep) and (lastchar == 'W'): # degrees W
                # make sure the rest is just numeric
                fval = float(val[:-1])
                return(pyferret.AXISTYPE_LONGITUDE, val.upper())
            elif (not istimestep) and (lastchar == 'N'): # degrees N
                # make sure the rest is just numeric
                fval = float(val[:-1])
                return(pyferret.AXISTYPE_LATITUDE, val.upper())
            elif (not istimestep) and (lastchar == 'S'): # degrees S
                # make sure the rest is just numeric
                fval = float(val[:-1])
                return(pyferret.AXISTYPE_LATITUDE, val.upper())
            elif (not istimestep) and (lastchar == 'M'): # meters (or kilometers, etc.)
                return(pyferret.AXISTYPE_LEVEL, val.upper())
            elif istimestep and (lastchar == 'Y'): # years
                fval = float(val[:-1])
                tlist = [ 0, 0, 0, 0, 0, 0 ]
                tlist[pyferret.TIMEARRAY_YEARINDEX] = fval
                return (pyferret.AXISTYPE_TIME, tlist)
            elif istimestep and (lastchar == 'D'): # days
                fval = float(val[:-1])
                tlist = [ 0, 0, 0, 0, 0, 0 ]
                tlist[pyferret.TIMEARRAY_DAYINDEX] = fval
                return (pyferret.AXISTYPE_TIME, tlist)
            elif istimestep and (lastchar == 'H'): # hours
                fval = float(val[:-1])
                tlist = [ 0, 0, 0, 0, 0, 0 ]
                tlist[pyferret.TIMEARRAY_HOURINDEX] = fval
                return (pyferret.AXISTYPE_TIME, tlist)
            elif istimestep and (lastchar == 'M'): # minutes
                fval = float(val[:-1])
                tlist = [ 0, 0, 0, 0, 0, 0 ]
                tlist[pyferret.TIMEARRAY_MINUTEINDEX] = fval
                return (pyferret.AXISTYPE_TIME, tlist)
            elif istimestep and (lastchar == 'S'): # seconds
                fval = float(val[:-1])
                tlist = [ 0, 0, 0, 0, 0, 0 ]
                tlist[pyferret.TIMEARRAY_SECONDINDEX] = fval
                return (pyferret.AXISTYPE_TIME, tlist)
            else:
                # maybe just numeric string; if not, will raise an exception
                fval = float(val)
                return(pyferret.AXISTYPE_ABSTRACT, fval)
        except Exception:
            raise ValueError('unable to parse: %s' % val)


    @staticmethod
    def _makedatestring(timearray):
        '''
        Creates a date and time string for the format DD-MON-YYYY HH:MM:SS 
        corresponding the values in the given time array.  If the year is 
        zero, -YYYY is omitted.  If the seconds is zero, :SS is omitted; 
        if hours, minutes, and seconds are all zero, HH:MM:SS is omitted.
            timearray: tuple of six int with time values given by the indices
                pyferret.TIMEARRAY_DAYINDEX
                pyferret.TIMEARRAY_MONTHINDEX
                pyferret.TIMEARRAY_YEARINDEX
                pyferret.TIMEARRAY_HOURINDEX
                pyferret.TIMEARRAY_MINUTEINDEX
                pyferret.TIMEARRAY_SECONDINDEX
        '''
        day = timearray[pyferret.TIMEARRAY_DAYINDEX]
        monthstr = pyferret.datamethods._UC_MONTH_NAMES[timearray[pyferret.TIMEARRAY_MONTHINDEX]]
        year = timearray[pyferret.TIMEARRAY_YEARINDEX]
        hour = timearray[pyferret.TIMEARRAY_HOURINDEX]
        minute = timearray[pyferret.TIMEARRAY_MINUTEINDEX]
        second = timearray[pyferret.TIMEARRAY_SECONDINDEX]
        if year > 0:
            datestr = '%02d-%3s-%04d' % (day, monthstr, year)
        else:
            datestr = '%02d-%3s' % (day, monthstr)
        if second > 0:
            timestr = ' %02d:%02d:%02d' % (hour, minute, second)
        elif (minute > 0) or (hour > 0):
            timestr = ' %02d:%02d' % (hour, minute)
        else:
            timestr = ''
        return datestr + timestr
 




"""
Performs a two-sided T-test that the provided sample
comes from a population with the given mean(s).
"""

from __future__ import print_function

import numpy
import pyferret
import scipy.stats


def ferret_init(id):
    """
    Initialization for the stats_ttest1 PyEF
    """
    axes_values = [ pyferret.AXIS_DOES_NOT_EXIST ] * pyferret.MAX_FERRET_NDIM
    axes_values[0] = pyferret.AXIS_CUSTOM
    axes_values[1] = pyferret.AXIS_CUSTOM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 2,
                "descript": "Returns [i=1] two-sided T-test stat. and [i=2] prob. " \
                            "for sample data coming from pop. with given mean(s).",
                "axes": axes_values,
                "argnames": ( "SAMPLE", "POPMEANS", ),
                "argdescripts": ( "Sample data to compare",
                                  "Proposed population means (averages)", ),
                "argtypes": ( pyferret.FLOAT_ARRAY, pyferret.FLOAT_ARRAY, ),
                "influences": ( false_influences, false_influences, ),
              }
    return retdict


def ferret_custom_axes(id):
    """
    Define custom axis of the stats_ttest1 Ferret PyEF
    """
    arglen = 1
    for axis in ( pyferret.X_AXIS, pyferret.Y_AXIS, pyferret.Z_AXIS, 
                  pyferret.T_AXIS, pyferret.E_AXIS, pyferret.F_AXIS ):
        axis_info = pyferret.get_axis_info(id, pyferret.ARG2, axis)
        num = axis_info.get("size", -1)
        if num > 0:
            arglen *= num
    # if all axes have undefined lengths, assume it is a single value
    axis_defs = [ None ] * pyferret.MAX_FERRET_NDIM
    axis_defs[0] = ( 1, 2, 1, "T,P", False )
    axis_defs[1] = (1, arglen, 1, "MEAN_INDEX", False)
    return axis_defs


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Performs a two-sided T-test that the provided sampe comes from a
    population with the given mean values.  The sample is given in
    inputs[0] and the mean values are given in inputs[1].  The test
    statistic value and two-tailed probability are returned in result
    along the first axis for each mean along the second axis.
    Undefined data in inputs[0] are removed before performing the test.
    """
    # get the valid sample values as 64-bit floats
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    values = numpy.array(inputs[0][goodmask], dtype=numpy.float64)
    # get the good mean values
    # need to flatten so the mask is one-dimensional
    means = inputs[1].reshape(-1)
    badmask = ( numpy.fabs(means - inpbdfs[1]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(means))
    goodmask = numpy.logical_not(badmask)
    means = numpy.array(means[goodmask], dtype=numpy.float64)
    # perform the test and assign the results
    fitparams = scipy.stats.ttest_1samp(values, means)
    result[:] = resbdf
    # T-test statistics
    result[0, goodmask, 0, 0, 0, 0] = fitparams[0]
    # probabilities
    result[1, goodmask, 0, 0, 0, 0] = fitparams[1]


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init does not have problems
    info = ferret_init(0)

    # Get a random sample from a normal distribution
    ydim = 200
    zdim = 150
    mu = 5.0
    sigma = 2.0
    sample = scipy.stats.norm(mu, sigma).rvs(ydim * zdim)
    means = numpy.linspace(mu - sigma, mu + sigma, 5)

    # setup for the call to ferret_compute
    inpbdfs = numpy.array([-9999.0, -8888.0], dtype=numpy.float64)
    resbdf  = numpy.array([-7777.0], dtype=numpy.float64)
    samparr = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    trimsamp = [ ]
    index = 0
    for j in range(ydim):
        for k in range(zdim):
            if (index % 71) == 3:
                samparr[0, j, k, 0, 0, 0] = inpbdfs[0]
            else:
                samparr[0, j, k, 0, 0, 0] = sample[index]
                trimsamp.append(sample[index])
            index += 1

    # computing each t-stat and p-val individually to check array arg return
    tvals = [ ]
    pvals = [ ]
    for m in means:
       (t, p) = scipy.stats.ttest_1samp(trimsamp, m)
       tvals.append(t)
       pvals.append(p)

    meanarr = inpbdfs[1] * numpy.ones((1, 2, 5, 1, 1, 1), dtype=numpy.float64, order='F')
    expect = resbdf * numpy.ones((2, 10, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    meanarr[0, 0, 1, 0, 0, 0] = means[0]
    expect[:, 1, 0, 0, 0, 0] = [ tvals[0], pvals[0] ]
    meanarr[0, 0, 3, 0, 0, 0] = mu - 0.5 * sigma
    expect[:, 3, 0, 0, 0, 0] = [ tvals[1], pvals[1] ]
    meanarr[0, 1, 0, 0, 0, 0] = mu
    expect[:, 5, 0, 0, 0, 0] = [ tvals[2], pvals[2] ]
    meanarr[0, 1, 2, 0, 0, 0] = mu + 0.5 * sigma
    expect[:, 7, 0, 0, 0, 0] = [ tvals[3], pvals[3] ]
    meanarr[0, 1, 4, 0, 0, 0] = mu + sigma
    expect[:, 9, 0, 0, 0, 0] = [ tvals[4], pvals[4] ]
    result = -6666.0 * numpy.ones((2, 10, 1, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute and check the results
    ferret_compute(0, result, resbdf, (samparr, meanarr), inpbdfs)
    if not numpy.allclose(result, expect):
        print("expect[:,:,0,0,0,0]:\n   %s" % str(expect[:, :, 0, 0, 0, 0]))
        print("result[:,:,0,0,0,0]:\n   %s" % str(result[:, :, 0, 0, 0, 0]))
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




"""
Returns histogram bin counts for a given array of values.
"""

from __future__ import print_function

import numpy
import pyferret

# The following is just to circumvent to call to pyferret.get_axis_info for testing
DOING_UNIT_TEST = False

def ferret_init(id):
    """
    Initialization for the stats_histogram Ferret PyEF
    """
    axes_values = [ pyferret.AXIS_IMPLIED_BY_ARGS ] * pyferret.MAX_FERRET_NDIM
    true_influences = [ True ] * pyferret.MAX_FERRET_NDIM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 2,
                "descript": "Returns unweighted histogram bin counts for all defined data values",
                "axes": axes_values,
                "argnames": ("VALS", "BINS_TEMPLATE"),
                "argdescripts": ("Values to put into bins and then count",
                                 "Template argument whose one defined axis gives midpoints of bins"),
                "argtypes": (pyferret.FLOAT_ARRAY, pyferret.FLOAT_ARRAY),
                "influences": (false_influences, true_influences),
              }
    return retdict


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Assigns result with histogram bin counts of data in inputs[0].  Bin
    limits are defined using the values of the one defined non-singular
    axis associated with inputs[1].  The argument inputs[1] is otherwise
    unused.  Undefined values in inputs[0] are eliminated before binning.
    """
    # get the box limits of the one defined non-singular axis of the second argument
    if DOING_UNIT_TEST:
        limits_func = my_get_box_limits
    else:
        limits_func = pyferret.get_axis_box_limits
    limits_tuple = None
    axis_used = None
    for axis_num in (pyferret.X_AXIS, pyferret.Y_AXIS, pyferret.Z_AXIS, 
                     pyferret.T_AXIS, pyferret.E_AXIS, pyferret.F_AXIS):
        this_tuple = limits_func(id, pyferret.ARG2, axis_num)
        if (this_tuple != None) and (len(this_tuple[0]) > 1):
            if limits_tuple != None:
               raise ValueError("BINS_TEMPLATE has more than one defined non-singular axis")
            limits_tuple = this_tuple
            axis_used = axis_num
    if limits_tuple is None:
        raise ValueError("BINS_TEMPLATE does not have a defined non-singular axis")
    # get the histogram bin limits from the axis box limits
    if not numpy.allclose(limits_tuple[0][1:], limits_tuple[1][:-1]):
        raise ValueError("Unexpected error: gaps exist between axis box limits")
    bin_edges = numpy.empty( ( len(limits_tuple[1]) + 1, ), dtype=numpy.float64)
    bin_edges[0] = limits_tuple[0][0]
    bin_edges[1:] = limits_tuple[1]
    # get the clean data as a flattened array
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    values = inputs[0][goodmask]
    # compute the histogram and assign the counts to result
    (hist, edges) = numpy.histogram(values, bins=bin_edges)
    if axis_used == pyferret.X_AXIS:
        result[:,0,0,0,0,0] = hist
    elif axis_used == pyferret.Y_AXIS:
        result[0,:,0,0,0,0] = hist
    elif axis_used == pyferret.Z_AXIS:
        result[0,0,:,0,0,0] = hist
    elif axis_used == pyferret.T_AXIS:
        result[0,0,0,:,0,0] = hist
    elif axis_used == pyferret.E_AXIS:
        result[0,0,0,0,:,0] = hist
    elif axis_used == pyferret.F_AXIS:
        result[0,0,0,0,0,:] = hist
    else:
        raise ValueError("Unexpected axis_used value: %d" % axis_used)


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    DOING_UNIT_TEST = True
    # create the my_get_box_limits function for testing
    def my_get_box_limits(id, argnum, axisnum):
        if id != 0:
            raise ValueError("Unexpected my_get_box_limits argnum; expected: 0, found: %d" % \
                              argnum)
        if argnum != pyferret.ARG2:
            raise ValueError("Unexpected my_get_box_limits argnum; expected: %d, found: %d" % \
                              (pyferret.ARG2, argnum))
        if axisnum != pyferret.Z_AXIS:
            return None
        limits = numpy.array([1.0, 2.0, 3.0, 4.0, 6.0, 9.0], dtype=numpy.float64)
        return (limits[:-1], limits[1:])
    # create the input values array with values on the edges and outside
    values = numpy.arange(0.0, 10.2, 0.1, dtype=numpy.float64).reshape((1,6,1,17,1,1), order='F')
    # create the expected results array
    expected = -1.0 * numpy.ones((1,1,5,1,1,1), dtype=numpy.float64, order='F')
    expected[0,0,:,0,0,0] = (10.0, 10.0, 10.0, 20.0, 31.0)
    # make sure no errors when ferret_init called
    info = ferret_init(0)
    # make the call to ferret_compute
    result = 999.0 * expected
    resbdf = numpy.array([-1.0], dtype=numpy.float64)
    inpbdfs = numpy.array([-1.0, -1.0], dtype=numpy.float64)
    ferret_compute(0, result, resbdf, (values, None), inpbdfs)
    # verify the results
    if not numpy.allclose(result, expected):
        raise ValueError("Unexpected results; expected:\n%s\nfound:\n%s" % (str(expected), str(result)))
    print("Success")

